"""The Oracle backend."""

from __future__ import annotations

import atexit
import contextlib
import re
import warnings
from functools import cached_property
from operator import itemgetter
from typing import TYPE_CHECKING, Any

import oracledb
import sqlglot as sg
import sqlglot.expressions as sge

import ibis
import ibis.common.exceptions as exc
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends.base.sqlglot import STAR, SQLGlotBackend
from ibis.backends.base.sqlglot.compiler import TRUE, C
from ibis.backends.oracle.compiler import OracleCompiler

if TYPE_CHECKING:
    from collections.abc import Iterable

    import pandas as pd
    import pyrrow as pa


class Backend(SQLGlotBackend):
    name = "oracle"
    compiler = OracleCompiler()

    @cached_property
    def version(self):
        matched = re.search(r"(\d+)\.(\d+)\.(\d+)", self.con.version)
        return ".".join(matched.groups())

    def do_connect(
        self,
        *,
        user: str,
        password: str,
        host: str = "localhost",
        port: int = 1521,
        database: str | None = None,
        sid: str | None = None,
        service_name: str | None = None,
        dsn: str | None = None,
        **_: Any,
    ) -> None:
        """Create an Ibis client using the passed connection parameters.

        Parameters
        ----------
        user
            Username
        password
            Password
        host
            Hostname
        port
            Port
        database
            Used as an Oracle service name if provided.
        sid
            Unique name of an Oracle Instance, used to construct a DSN if
            provided.
        service_name
            Oracle service name, used to construct a DSN if provided.  Only one
            of database and service_name should be provided.
        dsn
            An Oracle Data Source Name.  If provided, overrides all other
            connection arguments except username and password.

        """
        # SID: unique name of an INSTANCE running an oracle process (a single, identifiable machine)
        # service name: an ALIAS to one (or many) individual instances that can
        # be hotswapped without the client knowing / caring
        if dsn is not None and (
            database is not None or sid is not None or service_name is not None
        ):
            warnings.warn(
                "Oracle DSN provided, overriding additional provided connection arguments"
            )

        if service_name is not None and database is not None:
            raise exc.IbisInputError(
                "Values provided for both service_name and database. "
                "Both of these values map to an Oracle service_name, "
                "please provide only one of them."
            )

        if service_name is None and database is not None:
            service_name = database

        if dsn is None:
            dsn = oracledb.makedsn(host, port, service_name=service_name, sid=sid)

        # We set the statement cache size to 0 because Oracle will otherwise
        # attempt to reuse prepared statements even if the type of the bound variable
        # has changed.
        # This is apparently accepted behavior.
        # https://python-oracledb.readthedocs.io/en/latest/user_guide/appendix_b.html#statement-caching-in-thin-and-thick-modes
        self.con = oracledb.connect(dsn, user=user, password=password, stmtcachesize=0)

        # turn on autocommit
        # TODO: it would be great if this worked but it doesn't seem to do the trick
        # I had to hack in the commit lines to the compiler
        # self.con.autocommit = True

        # Set to ensure decimals come back as decimals
        oracledb.defaults.fetch_decimals = True

    def _from_url(self, url: str, **kwargs):
        return self.do_connect(user=url.username, password=url.password, dsn=url.host)

    @property
    def current_database(self) -> str:
        with self._safe_raw_sql(sg.select(STAR).from_("global_name")) as cur:
            [(database,)] = cur.fetchall()
        return database

    @contextlib.contextmanager
    def begin(self):
        con = self.con
        cur = con.cursor()
        try:
            yield cur
        except Exception:
            con.rollback()
            raise
        else:
            con.commit()
        finally:
            cur.close()

    @contextlib.contextmanager
    def _safe_raw_sql(self, *args, **kwargs):
        with contextlib.closing(self.raw_sql(*args, **kwargs)) as result:
            yield result

    def raw_sql(self, query: str | sg.Expression, **kwargs: Any) -> Any:
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.name)

        con = self.con
        cursor = con.cursor()

        try:
            cursor.execute(query, **kwargs)
        except Exception:
            con.rollback()
            cursor.close()
            raise
        else:
            con.commit()
            return cursor

    def list_tables(
        self, like: str | None = None, schema: str | None = None
    ) -> list[str]:
        """List the tables in the database.

        Parameters
        ----------
        like
            A pattern to use for listing tables.
        schema
            The schema to perform the list against.

        """
        conditions = [TRUE]

        if schema is None:
            schema = self.con.username.upper()
        conditions = C.owner.eq(sge.convert(schema.upper()))

        tables = (
            sg.select("table_name", "owner")
            .from_(sg.table("all_tables"))
            .distinct()
            .where(conditions)
        )
        views = (
            sg.select("view_name", "owner")
            .from_(sg.table("all_views"))
            .distinct()
            .where(conditions)
        )
        sql = tables.union(views).sql(self.name)

        with self._safe_raw_sql(sql) as cur:
            out = cur.fetchall()

        return self._filter_with_like(map(itemgetter(0), out), like)

    def list_schemas(
        self, like: str | None = None, database: str | None = None
    ) -> list[str]:
        if database is not None:
            raise exc.UnsupportedArgumentError(
                "No cross-database schema access in Oracle"
            )

        query = sg.select("username").from_("all_users").order_by("username")

        with self._safe_raw_sql(query) as con:
            schemata = list(map(itemgetter(0), con))

        return self._filter_with_like(schemata, like)

    def get_schema(
        self, name: str, schema: str | None = None, database: str | None = None
    ) -> sch.Schema:
        if schema is None:
            schema = self.con.username.upper()
        stmt = (
            sg.select(
                "column_name",
                "data_type",
                sg.column("nullable").eq(sge.convert("Y")).as_("nullable"),
            )
            .from_(sg.table("all_tab_columns"))
            .where(sg.column("table_name").eq(sge.convert(name)))
            .where(sg.column("owner").eq(sge.convert(schema)))
        )
        with self._safe_raw_sql(stmt) as cur:
            result = cur.fetchall()

        if not result:
            raise exc.IbisError(f"Table not found: {name!r}")

        type_mapper = self.compiler.type_mapper
        fields = {
            name: type_mapper.from_string(type_string, nullable=nullable)
            for name, type_string, nullable in result
        }

        return sch.Schema(fields)

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: ibis.Schema | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ):
        """Create a table in Oracle.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists

        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self._to_sqlglot(table)
        else:
            query = None

        column_defs = [
            sge.ColumnDef(
                this=sg.to_identifier(colname, quoted=self.compiler.quoted),
                kind=self.compiler.type_mapper.from_ibis(typ),
                constraints=(
                    None
                    if typ.nullable
                    else [sge.ColumnConstraint(kind=sge.NotNullColumnConstraint())]
                ),
            )
            for colname, typ in (schema or table.schema()).items()
        ]

        if overwrite:
            temp_name = util.gen_name(f"{self.name}_table")
        else:
            temp_name = name

        initial_table = sg.table(
            temp_name, catalog=database, quoted=self.compiler.quoted
        )
        target = sge.Schema(this=initial_table, expressions=column_defs)

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            properties=sge.Properties(expressions=properties),
        )

        # This is the same table as initial_table unless overwrite == True
        final_table = sg.table(name, catalog=database, quoted=self.compiler.quoted)
        with self._safe_raw_sql(create_stmt) as cur:
            if query is not None:
                insert_stmt = sge.Insert(this=initial_table, expression=query).sql(
                    self.name
                )
                cur.execute(insert_stmt)

            if overwrite:
                self.drop_table(
                    final_table.name, final_table.catalog, final_table.db, force=True
                )
                cur.execute(
                    f"ALTER TABLE IF EXISTS {initial_table.sql(self.name)} RENAME TO {final_table.sql(self.name)}"
                )

        if schema is None:
            return self.table(name, schema=database)

        # preserve the input schema if it was provided
        return ops.DatabaseTable(
            name, schema=schema, source=self, namespace=ops.Namespace(database=database)
        ).to_expr()

    def drop_table(
        self,
        name: str,
        database: str | None = None,
        schema: str | None = None,
        force: bool = False,
    ) -> None:
        table = sg.table(name, db=schema, catalog=database, quoted=self.compiler.quoted)

        with self.begin() as bind:
            # global temporary tables cannot be dropped without first truncating them
            #
            # https://stackoverflow.com/questions/32423397/force-oracle-drop-global-temp-table
            #
            # ignore DatabaseError exceptions because the table may not exist
            # because it's already been deleted
            with contextlib.suppress(oracledb.DatabaseError):
                bind.execute(f"TRUNCATE TABLE {table.sql(self.name)}")

        super().drop_table(name, database=database, schema=schema, force=force)

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        schema = op.schema

        # only register if we haven't already done so
        if (name := op.name) not in self.list_tables():
            quoted = self.compiler.quoted
            column_defs = [
                sge.ColumnDef(
                    this=sg.to_identifier(colname, quoted=quoted),
                    kind=self.compiler.type_mapper.from_ibis(typ),
                    constraints=(
                        None
                        if typ.nullable
                        else [
                            sg.exp.ColumnConstraint(
                                kind=sg.exp.NotNullColumnConstraint()
                            )
                        ]
                    ),
                )
                for colname, typ in schema.items()
            ]

            create_stmt = sge.Create(
                kind="TABLE",
                this=sg.exp.Schema(
                    this=sg.to_identifier(name, quoted=quoted), expressions=column_defs
                ),
                properties=sge.Properties(expressions=[sge.TemporaryProperty()]),
            ).sql(self.name, pretty=True)

            data = op.data.to_frame().itertuples(index=False)
            specs = ", ".join(f":{i}" for i, _ in enumerate(schema))
            table = sg.table(name, quoted=quoted).sql(self.name)
            insert_stmt = f"INSERT INTO {table} VALUES ({specs})"
            with self.begin() as cur:
                cur.execute(create_stmt)
                for row in data:
                    cur.execute(insert_stmt, row)

        atexit.register(self._clean_up_tmp_table, name)

    def _metadata(self, query: str) -> Iterable[tuple[str, dt.DataType]]:
        name = util.gen_name("oracle_metadata")
        dialect = self.name

        try:
            sg_expr = sg.parse_one(query, into=sg.exp.Table, dialect=dialect)
        except sg.ParseError:
            sg_expr = sg.parse_one(query, dialect=dialect)

        # If query is a table, adjust the query accordingly
        if isinstance(sg_expr, sg.exp.Table):
            sg_expr = sg.select(STAR).from_(sg_expr)

        # TODO(gforsyth): followup -- this should probably be made a default
        # transform for quoting backends
        def transformer(node):
            if isinstance(node, sg.exp.Table):
                return sg.table(node.name, quoted=True)
            elif isinstance(node, sg.exp.Column):
                return sg.column(col=node.name, quoted=True)
            return node

        sg_expr = sg_expr.transform(transformer)

        this = sg.table(name, quoted=True)
        create_view = sg.exp.Create(kind="VIEW", this=this, expression=sg_expr).sql(
            dialect
        )
        drop_view = sg.exp.Drop(kind="VIEW", this=this).sql(dialect)

        metadata_query = (
            sg.select(
                C.column_name,
                C.data_type,
                C.data_precision,
                C.data_scale,
                C.nullable.eq("Y"),
            )
            .from_("all_tab_columns")
            .where(C.table_name.eq(name))
            .order_by(C.column_id)
            .sql(dialect)
        )

        with self.begin() as con:
            con.execute(create_view)
            try:
                results = con.execute(metadata_query).fetchall()
            finally:
                # drop the view no matter what
                con.execute(drop_view)

        # TODO: hand all this off to the type mapper
        for name, type_string, precision, scale, nullable in results:
            # NUMBER(null, null) --> FLOAT
            # (null, null) --> from_string()
            if type_string == "NUMBER" and precision is None and scale is None:
                typ = dt.Float64(nullable=nullable)

            # (null, 0) --> INT
            # (null, 3), (null, 6), (null, 9) --> from_string() - TIMESTAMP(3)/(6)/(9)
            elif precision is None and (scale is not None and scale == 0):
                typ = dt.Int64(nullable=nullable)

            # NUMBER(*, 0) --> INT
            # (*, 0) --> from_string() - INTERVAL DAY(3) TO SECOND(0)
            elif (
                type_string == "NUMBER"
                and precision is not None
                and (scale is not None and scale == 0)
            ):
                typ = dt.Int64(nullable=nullable)

            # NUMBER(*, > 0) --> DECIMAL
            # (*, > 0) --> from_string() - INTERVAL DAY(3) TO SECOND(2)
            elif (
                type_string == "NUMBER"
                and precision is not None
                and (scale is not None and scale > 0)
            ):
                typ = dt.Decimal(precision=precision, scale=scale, nullable=nullable)

            else:
                typ = self.compiler.type_mapper.from_string(
                    type_string, nullable=nullable
                )
            yield name, typ

    def _fetch_from_cursor(self, cursor, schema: sch.Schema) -> pd.DataFrame:
        # TODO(gforsyth): this can probably be generalized a bit and put into
        # the base backend (or a mixin)
        import pandas as pd

        from ibis.backends.oracle.converter import OraclePandasData

        try:
            df = pd.DataFrame.from_records(
                cursor, columns=schema.names, coerce_float=True
            )
        except Exception:
            # clean up the cursor if we fail to create the DataFrame
            #
            # in the sqlite case failing to close the cursor results in
            # artificially locked tables
            cursor.close()
            raise
        df = OraclePandasData.convert_table(df, schema)
        return df

    def _clean_up_tmp_table(self, name: str) -> None:
        with self.begin() as bind:
            # global temporary tables cannot be dropped without first truncating them
            #
            # https://stackoverflow.com/questions/32423397/force-oracle-drop-global-temp-table
            #
            # ignore DatabaseError exceptions because the table may not exist
            # because it's already been deleted
            with contextlib.suppress(oracledb.DatabaseError):
                bind.execute(f'TRUNCATE TABLE "{name}"')
            with contextlib.suppress(oracledb.DatabaseError):
                bind.execute(f'DROP TABLE "{name}"')

    def _clean_up_cached_table(self, op):
        self._clean_up_tmp_table(op.name)



from __future__ import annotations

import pytest

import ibis
import ibis.common.exceptions as com
import ibis.expr.operations as ops
from ibis.expr.rewrites import simplify
from ibis.expr.tests.test_newrels import join_tables

# Place to collect esoteric expression analysis bugs and tests


def test_rewrite_join_projection_without_other_ops(con):
    # Star schema with fact table
    table = con.table("star1")
    table2 = con.table("star2")
    table3 = con.table("star3")

    filtered = table[table["f"] > 0]

    pred1 = table["foo_id"] == table2["foo_id"]
    pred2 = filtered["bar_id"] == table3["bar_id"]

    j1 = filtered.left_join(table2, [pred1])
    j2 = j1.inner_join(table3, [pred2])
    # Project out the desired fields
    view = j2[[filtered, table2["value1"], table3["value2"]]]

    with join_tables(filtered, table2, table3) as (r1, r2, r3):
        # Construct the thing we expect to obtain
        expected = ops.JoinChain(
            first=r1,
            rest=[
                ops.JoinLink(
                    how="left",
                    table=r2,
                    predicates=[r1["foo_id"] == r2["foo_id"]],
                ),
                ops.JoinLink(
                    how="inner",
                    table=r3,
                    predicates=[r1["bar_id"] == r3["bar_id"]],
                ),
            ],
            values={
                "c": r1.c,
                "f": r1.f,
                "foo_id": r1.foo_id,
                "bar_id": r1.bar_id,
                "value1": r2.value1,
                "value2": r3.value2,
            },
        )
        assert view.op() == expected


def test_multiple_join_deeper_reference():
    # Join predicates down the chain might reference one or more root
    # tables in the hierarchy.
    table1 = ibis.table({"key1": "string", "key2": "string", "value1": "double"})
    table2 = ibis.table({"key3": "string", "value2": "double"})
    table3 = ibis.table({"key4": "string", "value3": "double"})

    joined = table1.inner_join(table2, [table1["key1"] == table2["key3"]])
    joined2 = joined.inner_join(table3, [table1["key2"] == table3["key4"]])

    # it works, what more should we test here?
    repr(joined2)


def test_filter_on_projected_field(con):
    # See #173. Impala and other SQL engines do not allow filtering on a
    # just-created alias in a projection
    region = con.table("tpch_region")
    nation = con.table("tpch_nation")
    customer = con.table("tpch_customer")
    orders = con.table("tpch_orders")

    fields_of_interest = [
        customer,
        region.r_name.name("region"),
        orders.o_totalprice.name("amount"),
        orders.o_orderdate.cast("timestamp").name("odate"),
    ]

    all_join = (
        region.join(nation, region.r_regionkey == nation.n_regionkey)
        .join(customer, customer.c_nationkey == nation.n_nationkey)
        .join(orders, orders.o_custkey == customer.c_custkey)
    )

    tpch = all_join[fields_of_interest]

    # Correlated subquery, yikes!
    t2 = tpch.view()
    conditional_avg = t2[(t2.region == tpch.region)].amount.mean()

    # `amount` is part of the projection above as an aliased field
    amount_filter = tpch.amount > conditional_avg

    result = tpch.filter([amount_filter])

    # Now then! Predicate pushdown here is inappropriate, so we check that
    # it didn't occur.
    assert isinstance(result.op(), ops.Filter)
    assert result.op().parent == tpch.op()


def test_join_predicate_from_derived_raises():
    # Join predicate references a derived table, but we can salvage and
    # rewrite it to get the join semantics out
    # see ibis #74
    table = ibis.table([("c", "int32"), ("f", "double"), ("g", "string")], "foo_table")

    table2 = ibis.table([("key", "string"), ("value", "double")], "bar_table")

    filter_pred = table["f"] > 0
    table3 = table[filter_pred]

    with pytest.raises(com.IntegrityError, match="they belong to another relation"):
        # TODO(kszucs): could be smarter actually and rewrite the predicate
        # to contain the conditions from the filter
        table.inner_join(table2, [table3["g"] == table2["key"]])


def test_bad_join_predicate_raises():
    table = ibis.table([("c", "int32"), ("f", "double"), ("g", "string")], "foo_table")
    table2 = ibis.table([("key", "string"), ("value", "double")], "bar_table")
    table3 = ibis.table([("key", "string"), ("value", "double")], "baz_table")

    with pytest.raises(com.IntegrityError):
        table.inner_join(table2, [table["g"] == table3["key"]])


def test_filter_self_join():
    # GH #667
    purchases = ibis.table(
        [
            ("region", "string"),
            ("kind", "string"),
            ("user", "int64"),
            ("amount", "double"),
        ],
        "purchases",
    )

    metric = purchases.amount.sum().name("total")
    agged = purchases.group_by(["region", "kind"]).aggregate(metric)
    assert agged.op() == ops.Aggregate(
        parent=purchases,
        groups={"region": purchases.region, "kind": purchases.kind},
        metrics={"total": purchases.amount.sum()},
    )

    left = agged[agged.kind == "foo"]
    right = agged[agged.kind == "bar"]
    assert left.op() == ops.Filter(
        parent=agged,
        predicates=[agged.kind == "foo"],
    )
    assert right.op() == ops.Filter(
        parent=agged,
        predicates=[agged.kind == "bar"],
    )

    cond = left.region == right.region
    joined = left.join(right, cond)
    metric = (left.total - right.total).name("diff")
    what = [left.region, metric]
    projected = joined.select(what)

    with join_tables(left, right) as (r1, r2):
        join = ops.JoinChain(
            first=r1,
            rest=[
                ops.JoinLink("inner", r2, [r1.region == r2.region]),
            ],
            values={
                "region": r1.region,
                "diff": r1.total - r2.total,
            },
        )
        assert projected.op() == join


def test_is_ancestor_analytic():
    x = ibis.table(ibis.schema([("col", "int32")]), "x")
    with_filter_col = x[x.columns + [ibis.null().name("filter")]]
    filtered = with_filter_col[with_filter_col["filter"].isnull()]
    subquery = filtered[filtered.columns]

    with_analytic = subquery[subquery.columns + [subquery.count().name("analytic")]]

    assert not subquery.op().equals(with_analytic.op())


# Pr 2635
def test_mutation_fusion_no_overwrite():
    """Test fusion with chained mutation that doesn't overwrite existing
    columns."""
    t = ibis.table(ibis.schema([("col", "int32")]), "t")

    result = t
    result = result.mutate(col1=t["col"] + 1)
    result = result.mutate(col2=t["col"] + 2)
    result = result.mutate(col3=t["col"] + 3)

    simplified = simplify(result.op())
    assert simplified == ops.Project(
        parent=t,
        values={
            "col": t["col"],
            "col1": t["col"] + 1,
            "col2": t["col"] + 2,
            "col3": t["col"] + 3,
        },
    )


# Pr 2635
def test_mutation_fusion_overwrite():
    """Test fusion with chained mutation that overwrites existing columns."""
    t = ibis.table(ibis.schema([("col", "int32")]), "t")

    result = t

    result = result.mutate(col1=t["col"] + 1)
    result = result.mutate(col2=t["col"] + 2)
    result = result.mutate(col3=t["col"] + 3)
    result = result.mutate(col=t["col"] - 1)

    with pytest.raises(com.IntegrityError):
        # unable to dereference the column since result doesn't contain it anymore
        result.mutate(col4=t["col"] + 4)

    simplified = simplify(result.op())
    assert simplified == ops.Project(
        parent=t,
        values={
            "col": t["col"] - 1,
            "col1": t["col"] + 1,
            "col2": t["col"] + 2,
            "col3": t["col"] + 3,
        },
    )


# Pr 2635
def test_select_filter_mutate_fusion():
    """Test fusion with filter followed by mutation on the same input."""

    t = ibis.table(ibis.schema([("col", "float32")]), "t")

    t1 = t[["col"]]
    assert t1.op() == ops.Project(parent=t, values={"col": t.col})

    t2 = t1[t1["col"].isnan()]
    assert t2.op() == ops.Filter(parent=t1, predicates=[t1.col.isnan()])

    t3 = t2.mutate(col=t2["col"].cast("int32"))
    assert t3.op() == ops.Project(parent=t2, values={"col": t2.col.cast("int32")})

    # create the expected expression
    filt = ops.Filter(parent=t, predicates=[t.col.isnan()]).to_expr()
    proj = ops.Project(parent=filt, values={"col": filt.col.cast("int32")}).to_expr()

    t3_opt = simplify(t3.op()).to_expr()
    assert t3_opt.equals(proj)


def test_agg_selection_does_not_share_roots():
    t = ibis.table(dict(a="string"), name="t")
    s = ibis.table(dict(b="float64"), name="s")
    gb = t.group_by("a")
    n = s.count()

    with pytest.raises(com.IntegrityError, match=" they belong to another relation"):
        gb.aggregate(n=n)



from __future__ import annotations

import numbers  # noqa: TCH003
from typing import Literal

from public import public

import ibis.expr.datashape as ds
import ibis.expr.datatypes as dt
from ibis.common.annotations import ValidationError, attribute
from ibis.common.typing import VarTuple  # noqa: TCH001
from ibis.expr.operations.core import Column, Value


@public
class Bucket(Value):
    arg: Column[dt.Numeric | dt.Boolean]
    buckets: VarTuple[numbers.Real]
    closed: Literal["left", "right"] = "left"
    close_extreme: bool = True
    include_under: bool = False
    include_over: bool = False

    shape = ds.columnar

    @attribute
    def dtype(self):
        return dt.infer(self.nbuckets)

    def __init__(self, buckets, include_under, include_over, **kwargs):
        if not buckets:
            raise ValidationError("Must be at least one bucket edge")
        elif len(buckets) == 1:
            if not include_under or not include_over:
                raise ValidationError(
                    "If one bucket edge provided, must have "
                    "include_under=True and include_over=True"
                )
        super().__init__(
            buckets=buckets,
            include_under=include_under,
            include_over=include_over,
            **kwargs,
        )

    @property
    def nbuckets(self):
        return len(self.buckets) - 1 + self.include_over + self.include_under



from __future__ import annotations

import ibis

from .conftest import add_date, tpch_test


@tpch_test
def test_tpc_h05(customer, lineitem, orders, supplier, nation, region):
    """Local Supplier Volume Query (Q5)"""
    NAME = "ASIA"
    DATE = "1994-01-01"

    q = customer
    q = q.join(orders, customer.c_custkey == orders.o_custkey)
    q = q.join(lineitem, lineitem.l_orderkey == orders.o_orderkey)
    q = q.join(supplier, lineitem.l_suppkey == supplier.s_suppkey)
    q = q.join(
        nation,
        (customer.c_nationkey == supplier.s_nationkey)
        & (supplier.s_nationkey == nation.n_nationkey),
    )
    q = q.join(region, nation.n_regionkey == region.r_regionkey)

    q = q.filter(
        [
            q.r_name == NAME,
            q.o_orderdate >= ibis.date(DATE),
            q.o_orderdate < add_date(DATE, dy=1),
        ]
    )
    revexpr = q.l_extendedprice * (1 - q.l_discount)
    gq = q.group_by([q.n_name])
    q = gq.aggregate(revenue=revexpr.sum())
    q = q.order_by([ibis.desc(q.revenue)])
    return q



from __future__ import annotations

import subprocess
import sys
from importlib.metadata import EntryPoint
from typing import NamedTuple

import pytest

import ibis

# FIXME(kszucs): the following backends require the sqlite backend loaded
# def test_backends_are_cached():
#     assert ibis.sqlite is ibis.sqlite
#     del ibis.sqlite  # delete to force recreation
#     assert ibis.sqlite is ibis.sqlite


# def test_backends_tab_completion():
#     assert hasattr(ibis, "sqlite")
#     del ibis.sqlite  # delete to ensure not real attr
#     assert "sqlite" in dir(ibis)
#     assert ibis.sqlite is ibis.sqlite
#     assert "sqlite" in dir(ibis)  # in dir even if already created


# def test_public_backend_methods():
#     public = {m for m in dir(ibis.sqlite) if not m.startswith("_")}
#     assert public == {"connect", "compile", "has_operation", "add_operation", "name"}


def test_missing_backend():
    msg = "module 'ibis' has no attribute 'foo'."
    with pytest.raises(AttributeError, match=msg):
        ibis.foo  # noqa: B018


def test_multiple_backends(mocker):
    class Distribution(NamedTuple):
        entry_points: list[EntryPoint]

    entrypoints = [
        EntryPoint(
            name="foo",
            value="ibis.backends.backend1",
            group="ibis.backends",
        ),
        EntryPoint(
            name="foo",
            value="ibis.backends.backend2",
            group="ibis.backends",
        ),
    ]
    if sys.version_info < (3, 10):
        return_value = {"ibis.backends": entrypoints}
    else:
        return_value = entrypoints

    mocker.patch("importlib.metadata.entry_points", return_value=return_value)

    msg = r"\d+ packages found for backend 'foo'"
    with pytest.raises(RuntimeError, match=msg):
        ibis.foo  # noqa: B018


@pytest.mark.parametrize("module", ["pandas", "pyarrow"])
def test_no_import(module):
    script = f"""
import ibis
import sys

assert "{module}" not in sys.modules
"""
    subprocess.run([sys.executable, "-c", script], check=True)



from __future__ import annotations

import pytest

import ibis
from ibis.backends.impala.tests.conftest import translate


@pytest.fixture(scope="module")
def table(mockcon):
    return mockcon.table("functional_alltypes")


@pytest.mark.parametrize(
    "expr_fn",
    [
        pytest.param(
            lambda t: ibis.coalesce(t.string_col, "foo"), id="coalesce_scalar"
        ),
        pytest.param(
            lambda t: ibis.coalesce(t.int_col, t.bigint_col), id="coalesce_columns"
        ),
        pytest.param(
            lambda t: ibis.greatest(t.string_col, "foo"), id="greatest_scalar"
        ),
        pytest.param(
            lambda t: ibis.greatest(t.int_col, t.bigint_col), id="greatest_columns"
        ),
        pytest.param(lambda t: ibis.least(t.string_col, "foo"), id="least_scalar"),
        pytest.param(lambda t: ibis.least(t.int_col, t.bigint_col), id="least_columns"),
    ],
)
def test_varargs_functions(table, expr_fn, snapshot):
    t = table
    expr = expr_fn(t)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")



from __future__ import annotations

import pandas.testing as tm
import pytest
from pytest import param

import ibis.expr.datatypes as dt
from ibis import udf


@udf.scalar.builtin
def soundex(s: str) -> str:
    ...


@udf.scalar.builtin
def jarowinkler_similarity(a: str, b: str) -> float:
    ...


# TODO: allow multiple signatures
@udf.scalar.builtin(name="compress")
def compress_str(data: str, method: str) -> bytes:
    ...


@udf.scalar.builtin(name="compress")
def compress_bytes(data: bytes, method: str) -> bytes:
    ...


@pytest.mark.parametrize(
    ("func", "args"),
    [
        param(soundex, ("snow",), id="soundex"),
        param(jarowinkler_similarity, ("snow", "show"), id="jarowinkler_similarity"),
    ],
)
def test_builtin_scalar_udf(con, func, args):
    expr = func(*args)

    query = f"SELECT {func.__name__}({', '.join(map(repr, args))})"
    with con._safe_raw_sql(query) as cur:
        [(expected,)] = cur.fetchall()

    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ("func", "pyargs", "snowargs"),
    [
        param(compress_str, ("snow", "zstd"), ("'snow'", "'zstd'"), id="str"),
        param(compress_bytes, (b"snow", "zstd"), ("'snow'", "'zstd'"), id="bytes"),
    ],
)
def test_compress(con, func, pyargs, snowargs):
    expr = func(*pyargs)

    query = f"SELECT compress({', '.join(snowargs)})"
    with con._safe_raw_sql(query) as cur:
        [(expected,)] = cur.fetchall()

    assert con.execute(expr) == expected


@udf.agg.builtin
def minhash(x, y) -> dt.json:
    ...


@udf.agg.builtin
def approximate_jaccard_index(a) -> float:
    ...


def test_builtin_agg_udf(con):
    ft = con.tables.FUNCTIONAL_ALLTYPES.limit(2)
    ft = ft.select(mh=minhash(100, ft.string_col).over(group_by=ft.date_string_col))
    expr = ft.agg(aji=approximate_jaccard_index(ft.mh))

    result = expr.execute()
    query = """
    SELECT approximate_jaccard_index("mh") AS "aji"
    FROM (
        SELECT minhash(100, "string_col") OVER (PARTITION BY "date_string_col") AS "mh"
        FROM (
            SELECT * FROM "FUNCTIONAL_ALLTYPES" LIMIT 2
        )
    )
    """
    with con._safe_raw_sql(query) as cur:
        expected = cur.fetch_pandas_all()

    tm.assert_frame_equal(result, expected)


def test_xgboost_model(con):
    import ibis
    from ibis import _

    @udf.scalar.pandas(
        packages=("joblib", "xgboost"), imports=("@MODELS/model.joblib",)
    )
    def predict_price(
        carat_scaled: float, cut_encoded: int, color_encoded: int, clarity_encoded: int
    ) -> int:
        import sys

        import joblib
        import pandas as pd

        import_dir = sys._xoptions.get("snowflake_import_directory")
        model = joblib.load(f"{import_dir}model.joblib")
        df = pd.concat(
            [carat_scaled, cut_encoded, color_encoded, clarity_encoded], axis=1
        )
        df.columns = ["CARAT_SCALED", "CUT_ENCODED", "COLOR_ENCODED", "CLARITY_ENCODED"]
        return model.predict(df)

    def cases(value, mapping):
        """This should really be a top-level function or method."""
        expr = ibis.case()
        for k, v in mapping.items():
            expr = expr.when(value == k, v)
        return expr.end()

    diamonds = con.tables.DIAMONDS
    expr = diamonds.mutate(
        predicted_price=predict_price(
            (_.carat - _.carat.mean()) / _.carat.std(),
            cases(
                _.cut,
                {
                    c: i
                    for i, c in enumerate(
                        ("Fair", "Good", "Very Good", "Premium", "Ideal"), start=1
                    )
                },
            ),
            cases(_.color, {c: i for i, c in enumerate("DEFGHIJ", start=1)}),
            cases(
                _.clarity,
                {
                    c: i
                    for i, c in enumerate(
                        ("I1", "IF", "SI1", "SI2", "VS1", "VS2", "VVS1", "VVS2"),
                        start=1,
                    )
                },
            ),
        )
    )

    df = expr.execute()

    assert not df.empty
    assert "predicted_price" in df.columns
    assert len(df) == diamonds.count().execute()



from __future__ import annotations

import builtins
import sys
import tempfile

import pytest

from ibis.backends.bigquery.udf.core import PythonToJavaScriptTranslator, SymbolTable


def test_symbol_table():
    symbols = SymbolTable()
    assert symbols["a"] == "let a"
    assert symbols["a"] == "a"


def compile(f):
    return PythonToJavaScriptTranslator(f).compile()


def test_function_def(snapshot):
    def f(a, b):
        return a + b

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_variable_declaration(snapshot):
    def f():
        c = 1
        return c + 2

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_yield(snapshot):
    def f(a):
        yield from [1, 2, 3]

    js = compile(f)
    snapshot.assert_match(js, "out.js")


@pytest.mark.skipif(sys.platform == "win32", reason="Skip on Windows")
def test_yield_from(snapshot):
    d = {}

    with tempfile.NamedTemporaryFile("r+") as f:
        f.write(
            """\
def f(a):
    yield from [1, 2, 3]"""
        )
        f.seek(0)
        code = builtins.compile(f.read(), f.name, "exec")
        exec(code, d)
        f = d["f"]
        js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_assign(snapshot):
    def f():
        a = 1
        a = 2
        print(a)  # noqa: T201
        return 1

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def add(x, y):
    return x + y


def sub(x, y):
    return x - y


def mul(x, y):
    return x * y


def div(x, y):
    return x / y


@pytest.mark.parametrize("op", [add, sub, mul, div])
def test_binary_operators(op, snapshot):
    js = compile(op)
    snapshot.assert_match(js, "out.js")


def test_pow(snapshot):
    def f():
        a = 1
        return a**2

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_floordiv(snapshot):
    def f():
        a = 1
        return a // 2

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_unary_minus(snapshot):
    def f():
        a = 1
        return -a

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_unary_plus(snapshot):
    def f():
        a = 1
        return +a

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_true_false_none(snapshot):
    def f():
        a = True
        b = False
        c = None
        return a if c != None else b  # noqa: E711

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_str(snapshot):
    def f():
        a = "abc"
        return a

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_tuple(snapshot):
    def f():
        a = "a", "b", "c"
        return a

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_dict(snapshot):
    def f():
        a = {"a": 1, "b": 2}
        return a

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_continue(snapshot):
    def f():
        i = 0
        for i in [1, 2, 3]:
            if i == 1:
                continue
        return i

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_lambda_with_splat(snapshot):
    def f():
        def sum(sequence):
            total = 0
            for value in sequence:
                total += value
            return total

        splat_sum = lambda *args: sum(args)
        return splat_sum(1, 2, 3)

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_logical_not(snapshot):
    def f():
        a = True
        b = False
        return not a and not b

    js = compile(f)
    snapshot.assert_match(js, "out.js")


class Rectangle:
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def area(self):
        return self.width * self.height


class FancyRectangle(Rectangle):
    @property
    def perimeter(self):
        return self.width * 2 + self.height * 2


def test_class(snapshot):
    js = compile(Rectangle)
    snapshot.assert_match(js, "out.js")


def test_class_with_properties(snapshot):
    js = compile(FancyRectangle)
    snapshot.assert_match(js, "out.js")


def test_set_to_object(snapshot):
    def f(a):
        x = set()
        y = 1
        x.add(y)
        return y

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_setitem(snapshot):
    def f(a):
        x = {}
        y = "2"
        x[y] = y
        return x

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_delete(snapshot):
    def f(a):
        x = [a, 1, 2, 3]
        y = {"a": 1}
        del x[0], x[1]
        del x[0 + 3]
        del y.a
        return 1

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_scope_with_while(snapshot):
    def f():
        class Foo:
            def do_stuff(self):
                while True:
                    i = 1
                    i + 1
                    break

                while True:
                    i = 1
                    i + 1
                    break

                return 1

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_list_comp(snapshot):
    def f():
        x = [a + b for a, b in [(1, 2), (3, 4), (5, 6)] if a > 1 if b > 2]
        return x

    js = compile(f)
    snapshot.assert_match(js, "out.js")


@pytest.mark.xfail(raises=NotImplementedError, reason="Not yet implemented")
def test_nested_list_comp(snapshot):
    # TODO(phillipc): This can be done by nesting map calls followed by
    # N - 1 calls to array.reduce(Array.concat), where N is the number of
    # generators in the comprehension.
    def f():
        x = [
            a + b + c
            for a in [1, 4, 7]
            for b in [2, 5, 8]
            for c in [3, 6, 9]
            if a > 1
            if b > 2
            if c > 3
        ]
        return x

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_splat(snapshot):
    def f(x, y, z):
        def g(a, b, c):
            return a - b - c

        args = [x, y, z]
        return g(*args)

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_varargs(snapshot):
    def f(*args):
        return sum(*args)

    js = compile(f)
    snapshot.assert_match(js, "out.js")


def test_missing_vararg(snapshot):
    def my_range(n):
        return [1 for x in [n]]

    js = compile(my_range)
    snapshot.assert_match(js, "out.js")


def test_len_rewrite(snapshot):
    def my_func(a):
        return len(a)

    js = compile(my_func)
    snapshot.assert_match(js, "out.js")



from __future__ import annotations

from collections.abc import Mapping, Sequence

import pytest

from ibis.common.collections import frozendict
from ibis.common.graph import (
    Graph,
    Node,
    _coerce_finder,
    _coerce_replacer,
    _flatten_collections,
    _recursive_lookup,
    bfs,
    bfs_while,
    dfs,
    dfs_while,
)
from ibis.common.grounds import Annotable, Concrete
from ibis.common.patterns import Eq, If, InstanceOf, Object, TupleOf, _


class MyNode(Node):
    __match_args__ = ("name", "children")
    __slots__ = ("name", "children")

    def __init__(self, name, children):
        self.name = name
        self.children = children

    @property
    def __args__(self):
        return (self.name, self.children)

    @property
    def __argnames__(self):
        return ("name", "children")

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    def __hash__(self):
        return hash((self.__class__, self.name))

    def __eq__(self, other):
        return self.name == other.name

    def copy(self, name=None, children=None):
        return self.__class__(name or self.name, children or self.children)


C = MyNode(name="C", children=[])
D = MyNode(name="D", children=[])
E = MyNode(name="E", children=[])
B = MyNode(name="B", children=[D, E])
A = MyNode(name="A", children=[B, C])


def test_bfs():
    assert list(bfs(A).keys()) == [A, B, C, D, E]

    with pytest.raises(
        TypeError, match="must be an instance of ibis.common.graph.Node"
    ):
        bfs(1)


def test_construction():
    assert Graph(A) == bfs(A)


def test_graph_nodes():
    g = Graph(A)
    assert g.nodes() == {A, B, C, D, E}


def test_graph_repr():
    g = Graph(A)
    assert repr(g) == f"Graph({dict(g)})"


def test_dfs():
    assert list(dfs(A).keys()) == [D, E, B, C, A]

    with pytest.raises(
        TypeError, match="must be an instance of ibis.common.graph.Node"
    ):
        dfs(1)


def test_invert():
    g = dfs(A)
    assert g == {D: (), E: (), B: (D, E), C: (), A: (B, C)}

    i = g.invert()
    assert i == {D: (B,), E: (B,), B: (A,), C: (A,), A: ()}

    j = i.invert()
    assert j == g


def test_toposort():
    g, dependents = Graph(A).toposort()
    assert list(g.keys()) == [C, D, E, B, A]
    assert dependents == Graph(A).invert()


def test_toposort_cycle_detection():
    C = MyNode(name="C", children=[])
    A = MyNode(name="A", children=[C])
    B = MyNode(name="B", children=[A])
    A.children.append(B)

    # A depends on B which depends on A
    g = Graph(A)
    with pytest.raises(ValueError, match="cycle detected in the graph"):
        g.toposort()


def test_nested_children():
    a = MyNode(name="a", children=[])
    b = MyNode(name="b", children=[a])
    c = MyNode(name="c", children=[])
    d = MyNode(name="d", children=[])
    e = MyNode(name="e", children=[[b, c], {"d": d}])
    assert bfs(e) == {
        e: (b, c, d),
        b: (a,),
        c: (),
        d: (),
        a: (),
    }

    assert a.__children__ == ()
    assert b.__children__ == (a,)
    assert c.__children__ == ()
    assert d.__children__ == ()
    assert e.__children__ == (b, c, d)


@pytest.mark.parametrize("func", [bfs_while, dfs_while, Graph.from_bfs, Graph.from_dfs])
def test_traversals_with_filter(func):
    graph = func(A, filter=lambda x: x.name != "B")
    assert graph == {A: (C,), C: ()}

    graph = func(A, filter=lambda x: x.name != "D")
    assert graph == {E: (), B: (E,), A: (B, C), C: ()}


@pytest.mark.parametrize("func", [bfs_while, dfs_while, Graph.from_bfs, Graph.from_dfs])
def test_traversal_with_filtering_out_root(func):
    graph = func(A, filter=lambda x: x.name != "A")
    assert graph == {}


def test_replace_with_filtering_out_root():
    rule = InstanceOf(MyNode) >> MyNode(name="new", children=[])
    result = A.replace(rule, filter=If(lambda x: x.name != "A"))
    assert result == A


def test_replace_with_mapping():
    new_E = MyNode(name="e", children=[])
    new_D = MyNode(name="d", children=[])
    new_B = MyNode(name="B", children=[new_D, new_E])
    new_A = MyNode(name="A", children=[new_B, C])

    subs = {
        E: new_E,
        D: new_D,
    }
    result = A.replace(subs)
    assert result == new_A


def test_example():
    class Example(Annotable, Node):
        def __hash__(self):
            return hash((self.__class__, self.__args__))

    class Literal(Example):
        value = InstanceOf(object)

    class BoolLiteral(Literal):
        value = InstanceOf(bool)

    class And(Example):
        operands = TupleOf(InstanceOf(BoolLiteral))

    class Or(Example):
        operands = TupleOf(InstanceOf(BoolLiteral))

    class Collect(Example):
        arguments = TupleOf(TupleOf(InstanceOf(Example)) | InstanceOf(Example))

    a = BoolLiteral(True)
    b = BoolLiteral(False)
    c = BoolLiteral(True)
    d = BoolLiteral(False)

    and_ = And((a, b, c, d))
    or_ = Or((a, c))
    collect = Collect([and_, (or_, or_)])

    graph = bfs(collect)

    expected = {
        collect: (and_, or_, or_),
        or_: (a, c),
        and_: (a, b, c, d),
        a: (),
        b: (),
        # c and d are identical with a and b
    }
    assert graph == expected


def test_concrete_with_traversable_children():
    class Bool(Concrete, Node):
        pass

    class Value(Bool):
        value = InstanceOf(bool)

    class Either(Bool):
        left = InstanceOf(Bool)
        right = InstanceOf(Bool)

    class All(Bool):
        arguments = TupleOf(InstanceOf(Bool))
        strict = InstanceOf(bool)

    T, F = Value(True), Value(False)

    node = All((T, F), strict=True)
    assert node.__args__ == ((T, F), True)
    assert bfs(node) == {node: (T, F), T: (), F: ()}

    node = Either(T, F)
    assert node.__args__ == (T, F)
    assert bfs(node) == {node: (T, F), T: (), F: ()}

    node = All((T, Either(T, Either(T, F))), strict=False)
    assert node.__args__ == ((T, Either(T, Either(T, F))), False)
    assert bfs(node) == {
        node: (T, Either(T, Either(T, F))),
        T: (),
        F: (),
        Either(T, Either(T, F)): (T, Either(T, F)),
        Either(T, F): (T, F),
    }

    copied = node.copy(arguments=(T, F))
    assert copied == All((T, F), strict=False)


class MySequence(Sequence):
    def __init__(self, *items):
        self.items = items

    def __getitem__(self, index):
        raise AssertionError("must not be called")  # pragma: no cover

    def __len__(self):
        return len(self.items)


class MyMapping(Mapping):
    def __init__(self, **items):
        self.items = items

    def __getitem__(self, key):
        raise AssertionError("must not be called")  # pragma: no cover

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)


def test_flatten_collections():
    # test that flatten collections doesn't recurse into arbitrary mappings
    # and sequences, just the commonly used builtin ones: list, tuple, dict

    result = _flatten_collections([0.0, A, B, [C, D, (E, 6)], "7", MySequence(8, A)])
    assert list(result) == [A, B, C, D, E]

    result = _flatten_collections(
        {
            "a": 0.0,
            "b": A,
            "c": (MyMapping(d=B, e=3), frozendict(f=C)),
            "d": [5, "6", {"e": (D, 8.9)}],
        }
    )
    assert list(result) == [A, C, D]


def test_recursive_lookup():
    results = {A: "A", B: "B", C: "C", D: "D"}

    assert _recursive_lookup((A, B, "a", {"b": C}), results) == (
        "A",
        "B",
        "a",
        {"b": "C"},
    )
    assert _recursive_lookup({"a": B, "c": D}, results) == {
        "a": "B",
        "c": "D",
    }
    assert _recursive_lookup([A, B, "c"], results) == ("A", "B", "c")
    assert _recursive_lookup(A, results) == "A"

    my_seq = MySequence(A, "b", "c")
    my_map = MyMapping(a="a", b=B, c="c")
    assert _recursive_lookup((A, my_seq, [B, A], my_map), results) == (
        "A",
        my_seq,
        ("B", "A"),
        my_map,
    )


def test_coerce_finder():
    f = _coerce_finder(int)
    assert f(1) is True
    assert f("1") is False

    f = _coerce_finder((int, str))
    assert f(1) is True
    assert f("1") is True
    assert f(1.0) is False

    f = _coerce_finder(InstanceOf(bool))
    assert f(True) is True
    assert f(False) is True
    assert f(1) is False

    f = _coerce_finder(lambda x: x == 1)
    assert f(1) is True
    assert f(2) is False


def test_coerce_replacer():
    r = _coerce_replacer(lambda x, _, **kwargs: D)
    assert r(C, {}) == D

    r = _coerce_replacer({C: D, D: E})
    assert r(C, {}) == D
    assert r(D, {}) == E
    assert r(A, {}, name="A", children=[B, C]) == A

    r = _coerce_replacer(InstanceOf(MyNode) >> _.copy(name=_.name.lower()))
    assert r(C, {}, name="C", children=[]) == MyNode(name="c", children=[])
    assert r(D, {}, name="D", children=[]) == MyNode(name="d", children=[])


def test_node_find_using_type():
    class FooNode(MyNode):
        pass

    class BarNode(MyNode):
        pass

    C = BarNode(name="C", children=[])
    D = FooNode(name="D", children=[])
    E = BarNode(name="E", children=[])
    B = FooNode(name="B", children=[D, E])
    A = MyNode(name="A", children=[B, C])

    result = A.find(MyNode)
    assert result == [A, B, C, D, E]

    result = A.find(FooNode)
    assert result == [B, D]

    result = A.find(BarNode)
    assert result == [C, E]

    result = A.find((FooNode, BarNode))
    assert result == [B, C, D, E]

    result = A.find(int)
    assert result == []


def test_node_find_using_pattern():
    result = A.find(If(_.name == "C"))
    assert result == [C]

    result = A.find(Object(MyNode, name=Eq("D")))
    assert result == [D]

    result = A.find(If(_.children))
    assert result == [A, B]


def test_node_find_topmost_using_type():
    class FooNode(MyNode):
        pass

    G = FooNode(name="G", children=[A, B])
    E = MyNode(name="E", children=[G, G, A])

    result = E.find_topmost(FooNode)
    assert result == [G]

    result = E.find_topmost((FooNode, MyNode))
    assert result == [E]


def test_node_find_topmost_using_pattern():
    G = MyNode(name="G", children=[A, B])
    E = MyNode(name="E", children=[G, G, A])

    result = E.find_topmost(Object(MyNode, name="G") | Object(MyNode, name="B"))
    expected = [G, B]
    assert result == expected


def test_node_find_topmost_dont_traverse_the_same_node_twice():
    G = MyNode(name="G", children=[A, B])
    E = MyNode(name="E", children=[G, G, A])

    result = E.find_topmost(If(_.name == "G"))
    expected = [G]
    assert result == expected


def test_map_clear():
    Z = MyNode(name="Z", children=[A, A])
    Y = MyNode(name="Y", children=[A])
    X = MyNode(name="X", children=[Z, Y])
    result_sequence = {}

    def record_result_keys(node, results, **kwargs):
        result_sequence[node] = tuple(results.keys())
        return node

    expected_result_sequence = {
        C: (),
        D: (C,),
        E: (C, D),
        B: (C, D, E),
        A: (C, B),
        Z: (A,),
        Y: (A, Z),
        X: (Z, Y),
    }
    result = X.map_clear(record_result_keys)
    assert result == X
    assert result_sequence == expected_result_sequence



from __future__ import annotations

import numpy as np
import pandas as pd
import pandas.testing as tm
import pytest

import ibis

pytest.importorskip("pyspark")


@pytest.fixture
def t(con):
    return con.table("array_table")


@pytest.fixture
def df(con):
    return con._session.table("array_table").toPandas()


def test_array_length(t, df):
    result = t.mutate(length=t.array_int.length()).execute()
    expected = df.assign(length=df.array_int.map(lambda a: len(a)))
    tm.assert_frame_equal(result, expected)


def test_array_length_scalar(con):
    raw_value = [1, 2, 3]
    value = ibis.literal(raw_value)
    expr = value.length()
    result = con.execute(expr)
    expected = len(raw_value)
    assert result == expected


@pytest.mark.parametrize(
    ["start", "stop"],
    [
        (1, 3),
        (1, 1),
        (2, 3),
        (2, 5),
        (None, 3),
        (None, None),
        (3, None),
        (-3, None),
        (None, -3),
        (-3, -1),
    ],
)
def test_array_slice(t, df, start, stop):
    result = t.mutate(sliced=t.array_int[start:stop]).execute()
    expected = df.assign(sliced=df.array_int.map(lambda a: a[start:stop]))
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    ["start", "stop"],
    [
        (1, 3),
        (1, 1),
        (2, 3),
        (2, 5),
        (None, 3),
        (None, None),
        (3, None),
        (-3, None),
        (None, -3),
        (-3, -1),
    ],
)
def test_array_slice_scalar(con, start, stop):
    raw_value = [-11, 42, 10]
    value = ibis.literal(raw_value)
    expr = value[start:stop]
    result = con.execute(expr)
    expected = raw_value[start:stop]
    assert result == expected


@pytest.mark.parametrize("index", [1, 3, 4, 11, -11])
def test_array_index(t, df, index):
    expr = t[t.array_int[index].name("indexed")]
    result = expr.execute()

    expected = pd.DataFrame(
        {
            "indexed": df.array_int.apply(
                lambda x: x[index] if -len(x) <= index < len(x) else np.nan
            )
        }
    )
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize("index", [1, 3, 4, 11])
def test_array_index_scalar(con, index):
    raw_value = [-10, 1, 2, 42]
    value = ibis.literal(raw_value)
    expr = value[index]
    result = con.execute(expr)
    expected = raw_value[index] if index < len(raw_value) else np.nan
    assert result == expected or (np.isnan(result) and np.isnan(expected))


@pytest.mark.parametrize("op", [lambda x, y: x + y, lambda x, y: y + x])
def test_array_concat(t, df, op):
    x = t.array_int.cast("array<string>")
    y = t.array_str
    expr = op(x, y).name("array_result")
    result = expr.execute()

    expected = op(df.array_int.apply(lambda x: list(map(str, x))), df.array_str).rename(
        "array_result"
    )
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize("op", [lambda x, y: x + y, lambda x, y: y + x])
def test_array_concat_scalar(con, op):
    raw_left = [1, 2, 3]
    raw_right = [3, 4]
    left = ibis.literal(raw_left)
    right = ibis.literal(raw_right)
    expr = op(left, right)
    result = con.execute(expr)
    assert result == op(raw_left, raw_right)


@pytest.mark.parametrize("n", [1, 3, 4, 7, -2])  # negative returns empty list
@pytest.mark.parametrize("mul", [lambda x, n: x * n, lambda x, n: n * x])
def test_array_repeat(t, df, n, mul):
    expr = t.select(mul(t.array_int, n).name("repeated"))
    result = expr.execute()

    expected = pd.DataFrame({"repeated": df.array_int * n})
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize("n", [1, 3, 4, 7, -2])  # negative returns empty list
@pytest.mark.parametrize("mul", [lambda x, n: x * n, lambda x, n: n * x])
def test_array_repeat_scalar(con, n, mul):
    raw_array = [1, 2]
    array = ibis.literal(raw_array)
    expr = mul(array, n)
    result = con.execute(expr)
    expected = mul(raw_array, n)
    assert result == expected


def test_array_collect(t, df):
    expr = t.group_by(t.key).aggregate(collected=t.array_int.collect())
    result = expr.execute().sort_values("key").reset_index(drop=True)

    expected = (
        df.groupby("key")
        .array_int.apply(list)
        .reset_index()
        .rename(columns={"array_int": "collected"})
    )
    tm.assert_frame_equal(result, expected)


def test_array_filter(t, df):
    expr = t.select(t.array_int.filter(lambda item: item != 3).name("array_int"))
    result = expr.execute()

    df["array_int"] = df["array_int"].apply(
        lambda ar: [item for item in ar if item != 3]
    )
    expected = df[["array_int"]]
    tm.assert_frame_equal(result, expected)



from __future__ import annotations

import os
from contextlib import closing
from posixpath import join as pjoin

import pytest

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.types as ir
from ibis import util
from ibis.tests.util import assert_equal

pytest.importorskip("impala")

from impala.error import HiveServer2Error  # noqa: E402


@pytest.fixture
def temp_view(con) -> str:
    name = util.gen_name("view")
    yield name
    con.drop_view(name, force=True)


def test_create_exists_view(con, temp_view):
    assert temp_view not in con.list_tables()

    t1 = con.table("functional_alltypes").limit(1)
    t2 = con.create_view(temp_view, t1)

    assert temp_view in con.list_tables()
    assert not t2.execute().empty


def test_drop_non_empty_database(con, alltypes, temp_table):
    con.create_table(temp_table, alltypes)
    assert temp_table in con.list_tables()

    with pytest.raises(com.IntegrityError):
        con.drop_database("ibis_testing")


@pytest.fixture
def temp_db(con, tmp_dir):
    base = pjoin(tmp_dir, util.guid())
    name = util.gen_name("test_database")
    yield pjoin(base, name)
    con.drop_database(name)


def test_create_database_with_location(con, temp_db):
    name = os.path.basename(temp_db)
    con.create_database(name, path=temp_db)
    assert name in con.list_databases()


def test_create_table_with_location_execute(
    con, tmp_dir, alltypes, test_data_db, temp_table
):
    base = pjoin(tmp_dir, util.guid())
    name = f"test_{util.guid()}"
    tmp_path = pjoin(base, name)

    expr = alltypes

    con.create_table(temp_table, obj=expr, location=tmp_path, database=test_data_db)
    assert temp_table in con.list_tables()


def test_drop_table_not_exist(con):
    non_existent_table = f"ibis_table_{util.guid()}"
    with pytest.raises(HiveServer2Error):
        con.drop_table(non_existent_table)
    con.drop_table(non_existent_table, force=True)


def test_truncate_table(con, alltypes, temp_table):
    expr = alltypes.limit(1)

    con.create_table(temp_table, obj=expr)

    try:
        con.truncate_table(temp_table)
    except HiveServer2Error as e:
        if "AnalysisException" in e.args[0]:
            pytest.skip("TRUNCATE not available in this version of Impala")

    t = con.table(temp_table)
    nrows = t.count().execute()
    assert not nrows


def test_truncate_table_expression(con, alltypes, temp_table):
    expr = alltypes.limit(1)

    con.create_table(temp_table, obj=expr)
    t = con.table(temp_table)
    t.truncate()
    nrows = t.count().execute()
    assert not nrows


def test_ctas_from_table_expr(con, alltypes, temp_table):
    t = con.create_table(temp_table, alltypes)
    assert t.count().execute()


def test_create_empty_table(con, temp_table):
    schema = ibis.schema(
        [
            ("a", "string"),
            ("b", "timestamp"),
            ("c", "decimal(12, 8)"),
            ("d", "double"),
        ]
    )

    con.create_table(temp_table, schema=schema)

    result_schema = con.get_schema(temp_table)
    assert_equal(result_schema, schema)

    assert con.table(temp_table).execute().empty


def test_insert_table(con, alltypes, temp_table, test_data_db):
    expr = alltypes
    db = test_data_db

    con.create_table(temp_table, expr.limit(0), database=db)

    con.insert(temp_table, expr.limit(10), database=db)

    # check using ImpalaTable.insert
    t = con.table(temp_table, database=db)
    t.insert(expr.limit(10))

    sz = t.count()
    assert sz.execute() == 20

    # Overwrite and verify only 10 rows now
    t.insert(expr.limit(10), overwrite=True)
    assert sz.execute() == 10


def test_insert_validate_types(con, alltypes, test_data_db, temp_table):
    # GH #235
    db = test_data_db

    expr = alltypes
    con.create_table(
        temp_table,
        schema=expr["tinyint_col", "int_col", "string_col"].schema(),
        database=db,
    )

    t = con.table(temp_table, database=db)

    to_insert = expr[
        expr.tinyint_col, expr.smallint_col.name("int_col"), expr.string_col
    ]
    t.insert(to_insert.limit(10))

    to_insert = expr[
        expr.tinyint_col,
        expr.smallint_col.cast("int32").name("int_col"),
        expr.string_col,
    ]
    t.insert(to_insert.limit(10))

    to_insert = expr[expr.tinyint_col, expr.bigint_col.name("int_col"), expr.string_col]

    limit_expr = to_insert.limit(10)
    with pytest.raises(com.IbisError):
        t.insert(limit_expr)


def test_compute_stats(con):
    t = con.table("functional_alltypes")

    t.compute_stats()
    t.compute_stats(incremental=True)

    con.compute_stats("functional_alltypes")


@pytest.fixture
def created_view(con, alltypes):
    name = util.guid()
    expr = alltypes.limit(10)
    con.create_view(name, expr)
    return name


def test_drop_view(con, created_view):
    con.drop_view(created_view)
    assert created_view not in con.list_tables()


@pytest.fixture
def path_uuid():
    return f"change-location-{util.guid()}"


@pytest.fixture
def table(con, tmp_dir, path_uuid):
    table_name = f"table_{util.guid()}"
    fake_path = pjoin(tmp_dir, path_uuid)
    schema = ibis.schema([("foo", "string"), ("bar", "int64")])
    yield con.create_table(
        table_name, schema=schema, format="parquet", external=True, location=fake_path
    )
    con.drop_table(table_name)


def test_change_location(table, tmp_dir, path_uuid):
    old_loc = table.metadata().location

    new_path = pjoin(tmp_dir, "new-path")
    table.alter(location=new_path)

    new_loc = table.metadata().location
    assert new_loc == old_loc.replace(path_uuid, "new-path")


def test_change_properties(table):
    props = {"foo": "1", "bar": "2"}

    table.alter(tbl_properties=props)
    tbl_props = table.metadata().tbl_properties
    for k, v in props.items():
        assert v == tbl_props[k]

    table.alter(serde_properties=props)
    serde_props = table.metadata().serde_properties
    for k, v in props.items():
        assert v == serde_props[k]


def test_change_format(table):
    table.alter(format="avro")

    meta = table.metadata()
    assert "Avro" in meta.hive_format


def test_query_avro(con, test_data_dir):
    hdfs_path = pjoin(test_data_dir, "impala/avro/tpch/region")

    avro_schema = {
        "fields": [
            {"type": ["int", "null"], "name": "R_REGIONKEY"},
            {"type": ["string", "null"], "name": "R_NAME"},
            {"type": ["string", "null"], "name": "R_COMMENT"},
        ],
        "type": "record",
        "name": "a",
    }

    table = con.avro_file(hdfs_path, avro_schema)

    # table exists
    assert table._qualified_name in con.list_tables()

    expr = table.r_name.value_counts()
    expr.execute()

    assert table.count().execute() == 5

    df = table.execute()
    assert len(df) == 5


@pytest.fixture
def temp_table_id(con):
    name = "distinct"
    yield name
    con.drop_table(name)


def test_create_table_reserved_identifier(con, temp_table_id):
    expr = con.table("functional_alltypes")
    expected = expr.count().execute()
    con.create_table(temp_table_id, expr)
    result = con.table(temp_table_id).count().execute()
    assert result == expected


def test_query_delimited_file_directory(con, test_data_dir, temp_table):
    hdfs_path = pjoin(test_data_dir, "csv")

    schema = ibis.schema([("foo", "string"), ("bar", "double"), ("baz", "int8")])
    table = con.delimited_file(hdfs_path, schema, name=temp_table, delimiter=",")

    expr = (
        table[table.bar > 0]
        .group_by("foo")
        .aggregate(
            [
                table.bar.sum().name("sum(bar)"),
                table.baz.sum().name("mean(baz)"),
            ]
        )
    )
    assert expr.execute() is not None


@pytest.fixture
def temp_char_table(con):
    name = "testing_varchar_support"
    with closing(
        con.raw_sql(
            f"""\
CREATE TABLE IF NOT EXISTS {name} (
  group1 VARCHAR(10),
  group2 CHAR(10)
)"""
        )
    ):
        pass
    assert name in con.list_tables(), name
    yield con.table(name)
    con.drop_table(name, force=True)


def test_varchar_char_support(temp_char_table):
    assert isinstance(temp_char_table["group1"], ir.StringValue)
    assert isinstance(temp_char_table["group2"], ir.StringValue)


def test_access_kudu_table(kudu_table):
    assert kudu_table.columns == ["a"]
    assert kudu_table["a"].type() == dt.string



from __future__ import annotations

import datetime
from collections import OrderedDict

import numpy as np
import pandas as pd
import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
from ibis import _
from ibis.backends.tests.errors import OracleDatabaseError, PsycoPg2InternalError


@pytest.mark.parametrize(
    ("column", "raw_value"),
    [
        ("double_col", 0.0),
        ("double_col", 10.1),
        ("float_col", 1.1),
        ("float_col", 2.2),
    ],
)
def test_floating_scalar_parameter(backend, alltypes, df, column, raw_value):
    value = ibis.param(dt.double)
    expr = (alltypes[column] + value).name("tmp")
    expected = df[column] + raw_value
    result = expr.execute(params={value: raw_value})
    expected = backend.default_series_rename(expected)
    backend.assert_series_equal(result, expected, check_dtype=False)


@pytest.mark.parametrize(
    ("start_string", "end_string"),
    [("2009-03-01", "2010-07-03"), ("2014-12-01", "2017-01-05")],
)
@pytest.mark.notimpl(["trino", "druid"])
@pytest.mark.broken(["oracle"], raises=OracleDatabaseError)
def test_date_scalar_parameter(backend, alltypes, start_string, end_string):
    start, end = ibis.param(dt.date), ibis.param(dt.date)

    col = alltypes.timestamp_col.date()
    expr = col.between(start, end).name("output")
    expected_expr = col.between(start_string, end_string).name("output")

    result = expr.execute(params={start: start_string, end: end_string})
    expected = expected_expr.execute()

    backend.assert_series_equal(result, expected)


def test_timestamp_accepts_date_literals(alltypes):
    date_string = "2009-03-01"
    param = ibis.param(dt.timestamp)
    expr = alltypes.mutate(param=param)
    params = {param: date_string}
    assert expr.compile(params=params) is not None


@pytest.mark.notimpl(["impala", "druid", "oracle", "exasol"])
@pytest.mark.never(
    ["mysql", "sqlite", "mssql"], reason="backend will never implement array types"
)
def test_scalar_param_array(con):
    value = [1, 2, 3]
    param = ibis.param(dt.Array(dt.int64))
    result = con.execute(param.length().name("tmp"), params={param: value})
    assert result == len(value)


@pytest.mark.notimpl(
    [
        "datafusion",
        "impala",
        "postgres",
        "risingwave",
        "druid",
        "oracle",
        "exasol",
    ]
)
@pytest.mark.never(
    ["mysql", "sqlite", "mssql"],
    reason="mysql and sqlite will never implement struct types",
)
@pytest.mark.notimpl(["flink"], "WIP")
def test_scalar_param_struct(con):
    value = dict(a=1, b="abc", c=3.0)
    param = ibis.param("struct<a: int64, b: string, c: float64>")
    result = con.execute(param["a"], params={param: value})
    assert result == value["a"]


@pytest.mark.notimpl(["datafusion", "impala", "polars", "druid", "oracle", "exasol"])
@pytest.mark.never(
    ["mysql", "sqlite", "mssql"],
    reason="mysql and sqlite will never implement map types",
)
@pytest.mark.notyet(["bigquery"])
@pytest.mark.notimpl(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="function make_date(integer, integer, integer) does not exist",
)
def test_scalar_param_map(con):
    value = {"a": "ghi", "b": "def", "c": "abc"}
    param = ibis.param(dt.Map(dt.string, dt.string))
    result = con.execute(param["b"], params={param: value})
    assert result == value["b"]


@pytest.mark.parametrize(
    ("value", "dtype", "col"),
    [
        param("0", "string", "string_col", id="string"),
        param(0, "int64", "int_col", id="int"),
        param(0.0, "float64", "double_col", id="double"),
        param(
            True,
            "bool",
            "bool_col",
            id="bool",
            marks=[pytest.mark.notimpl(["druid"])],
        ),
        param(
            "2009-01-20 01:02:03",
            "timestamp",
            "timestamp_col",
            id="string_timestamp",
            marks=[pytest.mark.notimpl(["druid"])],
        ),
        param(
            datetime.date(2009, 1, 20),
            "timestamp",
            "timestamp_col",
            id="date_timestamp",
            marks=[pytest.mark.notimpl(["druid"])],
        ),
        param(
            datetime.datetime(2009, 1, 20, 1, 2, 3),
            "timestamp",
            "timestamp_col",
            id="datetime_timestamp",
            marks=[pytest.mark.notimpl(["druid"])],
        ),
    ],
)
def test_scalar_param(backend, alltypes, df, value, dtype, col):
    param = ibis.param(dtype)
    expr = alltypes.filter([_[col] == param])

    result = (
        expr.execute(params={param: value}).sort_values("id").reset_index(drop=True)
    )
    expected = df.loc[df[col] == value].sort_values("id").reset_index(drop=True)
    backend.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "value",
    ["2009-01-20", datetime.date(2009, 1, 20), datetime.datetime(2009, 1, 20)],
    ids=["string", "date", "datetime"],
)
@pytest.mark.notimpl(["druid", "oracle"])
def test_scalar_param_date(backend, alltypes, value):
    param = ibis.param("date")
    ds_col = alltypes.date_string_col
    month = ds_col[:2]
    day = ds_col[3:5]
    year = "20" + ds_col[6:8]
    date_col = (year + "-" + month + "-" + day).cast(param.type())

    base = alltypes.mutate(date_col=date_col)
    expr = (
        alltypes.mutate(date_col=date_col)
        .filter(lambda t: t.date_col == param)
        .drop("date_col")
    )

    result = (
        expr.execute(params={param: value}).sort_values("id").reset_index(drop=True)
    )
    df = base.execute()
    expected = (
        df.loc[
            pd.to_datetime(df.date_col).dt.normalize().dt.date
            == pd.Timestamp(value).normalize().date()
        ]
        .sort_values("id")
        .reset_index(drop=True)
        .drop(columns=["date_col"])
    )
    backend.assert_frame_equal(result, expected)


@pytest.mark.notyet(["flink", "mysql"], reason="no struct support")
@pytest.mark.notimpl(
    [
        "postgres",
        "risingwave",
        "datafusion",
        "clickhouse",
        "polars",
        "sqlite",
        "impala",
        "oracle",
        "mssql",
        "druid",
        "exasol",
    ]
)
def test_scalar_param_nested(con):
    param = ibis.param("struct<x: array<struct<y: array<double>>>>")
    value = OrderedDict([("x", [OrderedDict([("y", [1.0, 2.0, 3.0])])])])
    result = con.execute(param, {param: value})
    assert pytest.approx(result["x"][0]["y"]) == np.array([1.0, 2.0, 3.0])



import ibis


lit = ibis.timestamp("2018-01-01 00:00:00")
s = ibis.table(name="s", schema={"b": "string"})
t = ibis.table(name="t", schema={"a": "int64", "b": "string", "c": "timestamp"})
f = t.filter(t.c == lit)
joinchain = (
    f.select(f.a, f.b, lit.name("the_date"))
    .inner_join(s, f.select(f.a, f.b, lit.name("the_date")).b == s.b)
    .select(f.select(f.a, f.b, lit.name("the_date")).a)
)

result = joinchain.filter(joinchain.a < 1.0)



from __future__ import annotations

import inspect
from typing import Annotated, Union

import pytest

from ibis.common.annotations import (
    Argument,
    Attribute,
    Parameter,
    Signature,
    ValidationError,
    annotated,
    argument,
    attribute,
    optional,
)
from ibis.common.patterns import (
    Any,
    CoercedTo,
    InstanceOf,
    NoMatch,
    Option,
    TupleOf,
    pattern,
)

is_int = InstanceOf(int)


def test_argument_factory():
    a = argument(is_int, default=1, typehint=int)
    assert a == Argument(is_int, default=1, typehint=int)

    a = argument(is_int, default=1)
    assert a == Argument(is_int, default=1)

    a = argument(is_int)
    assert a == Argument(is_int)


def test_attribute_factory():
    a = attribute(is_int, default=1)
    assert a == Attribute(is_int, default=1)

    a = attribute(is_int)
    assert a == Attribute(is_int)

    a = attribute(default=2)
    assert a == Attribute(default=2)

    a = attribute(int, default=2)
    assert a == Attribute(int, default=2)


def test_annotations_are_immutable():
    a = argument(is_int, default=1)
    with pytest.raises(AttributeError):
        a.pattern = Any()
    with pytest.raises(AttributeError):
        a.default = 2

    a = attribute(is_int, default=1)
    with pytest.raises(AttributeError):
        a.pattern = Any()
    with pytest.raises(AttributeError):
        a.default = 2


def test_annotations_are_not_hashable():
    # in order to use the with mutable defaults
    a = argument(is_int, default=1)
    with pytest.raises(TypeError, match="unhashable type: 'Argument'"):
        hash(a)

    a = attribute(is_int, default=1)
    with pytest.raises(TypeError, match="unhashable type: 'Attribute'"):
        hash(a)


def test_argument_repr():
    argument = Argument(is_int, typehint=int, default=None)
    assert repr(argument) == (
        "Argument(pattern=InstanceOf(type=<class 'int'>), default=None, "
        "typehint=<class 'int'>, kind=<_ParameterKind.POSITIONAL_OR_KEYWORD: 1>)"
    )


def test_default_argument():
    annotation = Argument(pattern=lambda x, context: int(x), default=3)
    assert annotation.pattern.match(1, {}) == 1


@pytest.mark.parametrize(
    ("default", "expected"),
    [(None, None), (0, 0), ("default", "default")],
)
def test_optional_argument(default, expected):
    annotation = optional(default=default)
    assert annotation.pattern.match(None, {}) == expected


@pytest.mark.parametrize(
    ("argument", "value", "expected"),
    [
        (optional(Any(), default=None), None, None),
        (optional(Any(), default=None), "three", "three"),
        (optional(Any(), default=1), None, 1),
        (optional(CoercedTo(int), default=11), None, 11),
        (optional(CoercedTo(int), default=None), None, None),
        (optional(CoercedTo(int), default=None), 18, 18),
        (optional(CoercedTo(str), default=None), "caracal", "caracal"),
    ],
)
def test_valid_optional(argument, value, expected):
    assert argument.pattern.match(value, {}) == expected


def test_attribute_default_value():
    class Foo:
        a = 10

    assert not Attribute().has_default()

    field = Attribute(default=lambda self: self.a + 10)
    assert field.has_default()
    assert field == field

    assert field.get_default("b", Foo) == 20

    field2 = Attribute(pattern=lambda x, this: str(x), default=lambda self: self.a)
    assert field2.has_default()
    assert field != field2
    assert field2.get_default("b", Foo) == "10"


def test_parameter():
    def fn(x, this):
        return int(x) + this["other"]

    annot = argument(fn)
    p = Parameter("test", annotation=annot)

    assert p.annotation is annot
    assert p.default is inspect.Parameter.empty
    assert p.annotation.pattern.match("2", {"other": 1}) == 3

    ofn = optional(fn)
    op = Parameter("test", annotation=ofn)
    assert op.annotation.pattern == Option(fn, default=None)
    assert op.default is None
    assert op.annotation.pattern.match(None, {"other": 1}) is None

    with pytest.raises(TypeError, match="annotation must be an instance of Argument"):
        Parameter("wrong", annotation=Attribute(lambda x, context: x))


def test_signature():
    def to_int(x, this):
        return int(x)

    def add_other(x, this):
        return int(x) + this["other"]

    other = Parameter("other", annotation=Argument(to_int))
    this = Parameter("this", annotation=Argument(add_other))

    sig = Signature(parameters=[other, this])
    assert sig.validate(None, args=(1, 2), kwargs={}) == {"other": 1, "this": 3}
    assert sig.validate(None, args=(), kwargs=dict(other=1, this=2)) == {
        "other": 1,
        "this": 3,
    }
    assert sig.validate(None, args=(), kwargs=dict(this=2, other=1)) == {
        "other": 1,
        "this": 3,
    }


def test_signature_from_callable():
    def test(a: int, b: int, c: int = 1):
        ...

    sig = Signature.from_callable(test)
    assert sig.validate(test, args=(2, 3), kwargs={}) == {"a": 2, "b": 3, "c": 1}

    with pytest.raises(ValidationError):
        sig.validate(test, args=(2, 3, "4"), kwargs={})

    args, kwargs = sig.unbind(sig.validate(test, args=(2, 3), kwargs={}))
    assert args == (2, 3, 1)
    assert kwargs == {}


def test_signature_from_callable_with_varargs():
    def test(a: int, b: int, *args: int):
        ...

    sig = Signature.from_callable(test)
    assert sig.validate(test, args=(2, 3), kwargs={}) == {"a": 2, "b": 3, "args": ()}
    assert sig.validate(test, args=(2, 3, 4), kwargs={}) == {
        "a": 2,
        "b": 3,
        "args": (4,),
    }
    assert sig.validate(test, args=(2, 3, 4, 5), kwargs={}) == {
        "a": 2,
        "b": 3,
        "args": (4, 5),
    }
    assert sig.parameters["a"].annotation.typehint is int
    assert sig.parameters["b"].annotation.typehint is int
    assert sig.parameters["args"].annotation.typehint is int

    with pytest.raises(ValidationError):
        sig.validate(test, args=(2, 3, 4, "5"), kwargs={})

    args, kwargs = sig.unbind(sig.validate(test, args=(2, 3, 4, 5), kwargs={}))
    assert args == (2, 3, 4, 5)
    assert kwargs == {}


def test_signature_from_callable_with_positional_only_arguments(snapshot):
    def test(a: int, b: int, /, c: int = 1):
        ...

    sig = Signature.from_callable(test)
    assert sig.validate(test, args=(2, 3), kwargs={}) == {"a": 2, "b": 3, "c": 1}
    assert sig.validate(test, args=(2, 3, 4), kwargs={}) == {"a": 2, "b": 3, "c": 4}
    assert sig.validate(test, args=(2, 3), kwargs=dict(c=4)) == {"a": 2, "b": 3, "c": 4}

    with pytest.raises(ValidationError) as excinfo:
        sig.validate(test, args=(1,), kwargs=dict(b=2))
    snapshot.assert_match(str(excinfo.value), "parameter_is_positional_only.txt")

    args, kwargs = sig.unbind(sig.validate(test, args=(2, 3), kwargs={}))
    assert args == (2, 3, 1)
    assert kwargs == {}


def test_signature_from_callable_with_keyword_only_arguments(snapshot):
    def test(a: int, b: int, *, c: float, d: float = 0.0):
        ...

    sig = Signature.from_callable(test)
    assert sig.validate(test, args=(2, 3), kwargs=dict(c=4.0)) == {
        "a": 2,
        "b": 3,
        "c": 4.0,
        "d": 0.0,
    }
    assert sig.validate(test, args=(2, 3), kwargs=dict(c=4.0, d=5.0)) == {
        "a": 2,
        "b": 3,
        "c": 4.0,
        "d": 5.0,
    }

    with pytest.raises(ValidationError) as excinfo:
        sig.validate(test, args=(2, 3), kwargs={})
    snapshot.assert_match(str(excinfo.value), "missing_a_required_argument.txt")

    with pytest.raises(ValidationError) as excinfo:
        sig.validate(test, args=(2, 3, 4), kwargs={})
    snapshot.assert_match(str(excinfo.value), "too_many_positional_arguments.txt")

    args, kwargs = sig.unbind(sig.validate(test, args=(2, 3), kwargs=dict(c=4.0)))
    assert args == (2, 3)
    assert kwargs == {"c": 4.0, "d": 0.0}


def test_signature_unbind():
    def to_int(x, this):
        return int(x)

    def add_other(x, this):
        return int(x) + this["other"]

    other = Parameter("other", annotation=Argument(to_int))
    this = Parameter("this", annotation=Argument(add_other))

    sig = Signature(parameters=[other, this])
    params = sig.validate(None, args=(1,), kwargs=dict(this=2))

    args, kwargs = sig.unbind(params)
    assert args == (1, 3)
    assert kwargs == {}


a = Parameter("a", annotation=Argument(CoercedTo(float)))
b = Parameter("b", annotation=Argument(CoercedTo(float)))
c = Parameter("c", annotation=Argument(CoercedTo(float), default=0))
d = Parameter(
    "d",
    annotation=Argument(TupleOf(CoercedTo(float)), default=()),
)
e = Parameter("e", annotation=Argument(Option(CoercedTo(float)), default=None))
sig = Signature(parameters=[a, b, c, d, e])


@pytest.mark.parametrize("d", [(), (5, 6, 7)])
def test_signature_unbind_with_empty_variadic(d):
    params = sig.validate(None, args=(1, 2, 3, d), kwargs=dict(e=4))
    assert params == {"a": 1.0, "b": 2.0, "c": 3.0, "d": d, "e": 4.0}

    args, kwargs = sig.unbind(params)
    assert args == (1.0, 2.0, 3.0, tuple(map(float, d)), 4.0)
    assert kwargs == {}

    params_again = sig.validate(None, args=args, kwargs=kwargs)
    assert params_again == params


def test_annotated_function():
    @annotated(a=InstanceOf(int), b=InstanceOf(int), c=InstanceOf(int))
    def test(a, b, c=1):
        return a + b + c

    assert test(2, 3) == 6
    assert test(2, 3, 4) == 9
    assert test(2, 3, c=4) == 9
    assert test(a=2, b=3, c=4) == 9

    with pytest.raises(ValidationError):
        test(2, 3, c="4")

    @annotated(a=InstanceOf(int))
    def test(a, b, c=1):
        return (a, b, c)

    assert test(2, "3") == (2, "3", 1)


def test_annotated_function_with_type_annotations():
    @annotated()
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    assert test(2, 3) == 6

    @annotated
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    assert test(2, 3) == 6

    @annotated
    def test(a: int, b, c=1):
        return (a, b, c)

    assert test(2, 3, "4") == (2, 3, "4")


def test_annotated_function_with_return_type_annotation():
    @annotated
    def test_ok(a: int, b: int, c: int = 1) -> int:
        return a + b + c

    @annotated
    def test_wrong(a: int, b: int, c: int = 1) -> int:
        return "invalid result"

    assert test_ok(2, 3) == 6
    with pytest.raises(ValidationError):
        test_wrong(2, 3)


def test_annotated_function_with_keyword_overrides():
    @annotated(b=InstanceOf(float))
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    with pytest.raises(ValidationError):
        test(2, 3)

    assert test(2, 3.0) == 6.0


def test_annotated_function_with_list_overrides():
    @annotated([InstanceOf(int), InstanceOf(int), InstanceOf(float)])
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    with pytest.raises(ValidationError):
        test(2, 3, 4)


def test_annotated_function_with_list_overrides_and_return_override():
    @annotated([InstanceOf(int), InstanceOf(int), InstanceOf(float)], InstanceOf(float))
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    with pytest.raises(ValidationError):
        test(2, 3, 4)

    assert test(2, 3, 4.0) == 9.0


@pattern
def short_str(x, this):
    if len(x) > 3:
        return x
    else:
        return NoMatch


@pattern
def endswith_d(x, this):
    if x.endswith("d"):
        return x
    else:
        return NoMatch


def test_annotated_function_with_complex_type_annotations():
    @annotated
    def test(a: Annotated[str, short_str, endswith_d], b: Union[int, float]):
        return a, b

    assert test("abcd", 1) == ("abcd", 1)
    assert test("---d", 1.0) == ("---d", 1.0)

    with pytest.raises(ValidationError):
        test("---c", 1)
    with pytest.raises(ValidationError):
        test("123", 1)
    with pytest.raises(ValidationError):
        test("abcd", "qweqwe")


def test_annotated_function_without_annotations():
    @annotated
    def test(a, b, c):
        return a, b, c

    assert test(1, 2, 3) == (1, 2, 3)
    assert test.__signature__.parameters.keys() == {"a", "b", "c"}


def test_annotated_function_without_decoration(snapshot):
    def test(a, b, c):
        return a + b + c

    func = annotated(test)
    with pytest.raises(ValidationError) as excinfo:
        func(1, 2)
    snapshot.assert_match(str(excinfo.value), "error.txt")

    assert func(1, 2, c=3) == 6


def test_annotated_function_with_varargs():
    @annotated
    def test(a: float, b: float, *args: int):
        return sum((a, b) + args)

    assert test(1.0, 2.0, 3, 4) == 10.0
    assert test(1.0, 2.0, 3, 4, 5) == 15.0

    with pytest.raises(ValidationError):
        test(1.0, 2.0, 3, 4, 5, 6.0)


def test_annotated_function_with_varkwargs():
    @annotated
    def test(a: float, b: float, **kwargs: int):
        return sum((a, b) + tuple(kwargs.values()))

    assert test(1.0, 2.0, c=3, d=4) == 10.0
    assert test(1.0, 2.0, c=3, d=4, e=5) == 15.0

    with pytest.raises(ValidationError):
        test(1.0, 2.0, c=3, d=4, e=5, f=6.0)


def test_multiple_validation_failures():
    @annotated
    def test(a: float, b: float, *args: int, **kwargs: int):
        ...

    with pytest.raises(ValidationError) as excinfo:
        test(1.0, 2.0, 3.0, 4, c=5.0, d=6)

    assert len(excinfo.value.errors) == 2



from __future__ import annotations

import pytest
from pytest import param

import ibis.expr.datatypes as dt
from ibis.backends.flink.datatypes import FlinkType


@pytest.mark.parametrize(
    ("datatype", "expected"),
    [
        param(dt.float32, "FLOAT", id="float32"),
        param(dt.float64, "DOUBLE", id="float64"),
        param(dt.uint8, "TINYINT", id="uint8"),
        param(dt.uint16, "SMALLINT", id="uint16"),
        param(dt.uint32, "INT", id="uint32"),
        param(dt.int8, "TINYINT", id="int8"),
        param(dt.int16, "SMALLINT", id="int16"),
        param(dt.int32, "INT", id="int32"),
        param(dt.int64, "BIGINT", id="int64"),
        param(dt.string, "VARCHAR", id="string"),
        param(dt.date, "DATE", id="date"),
        param(dt.timestamp, "TIMESTAMP", id="datetime"),
        param(
            dt.Timestamp(timezone="UTC"),
            "TIMESTAMP",
            id="timestamp_with_utc_tz",
        ),
        param(
            dt.Timestamp(timezone="US/Eastern"),
            "TIMESTAMP",
            id="timestamp_with_other_tz",
        ),
    ],
)
def test_simple(datatype, expected):
    assert FlinkType.to_string(datatype) == expected



from __future__ import annotations

import hypothesis as h
import hypothesis.strategies as st
import numpy as np
import pytest

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
import ibis.expr.types as ir
import ibis.tests.strategies as its
from ibis.common.annotations import ValidationError


@h.given(its.null_dtype)
def test_null_dtype(dtype):
    assert isinstance(dtype, dt.Null)
    assert dtype.is_null() is True
    assert dtype.nullable is True


@h.given(its.boolean_dtype())
def test_boolean_dtype(dtype):
    assert isinstance(dtype, dt.Boolean)
    assert dtype.is_boolean() is True


@h.given(its.signed_integer_dtypes())
def test_signed_integer_dtype(dtype):
    assert isinstance(dtype, dt.SignedInteger)
    assert dtype.is_integer() is True


@h.given(its.unsigned_integer_dtypes())
def test_unsigned_integer_dtype(dtype):
    assert isinstance(dtype, dt.UnsignedInteger)
    assert dtype.is_integer() is True


@h.given(its.floating_dtypes())
def test_floating_dtype(dtype):
    assert isinstance(dtype, dt.Floating)
    assert dtype.is_floating() is True


@h.given(its.numeric_dtypes())
def test_numeric_dtype(dtype):
    assert isinstance(dtype, dt.Numeric)
    assert dtype.is_numeric() is True


@h.given(its.numeric_dtypes(nullable=st.just(True)))
def test_numeric_dtypes_nullable(dtype):
    assert dtype.nullable is True
    assert dtype.is_numeric() is True


@h.given(its.numeric_dtypes(nullable=st.just(False)))
def test_numeric_dtypes_non_nullable(dtype):
    assert dtype.nullable is False
    assert dtype.is_numeric() is True


@h.given(its.timestamp_dtype())
def test_timestamp_dtype(dtype):
    assert isinstance(dtype, dt.Timestamp)
    assert isinstance(dtype.timezone, (type(None), str))
    assert dtype.is_timestamp() is True


@h.given(its.interval_dtype())
def test_interval_dtype(dtype):
    assert isinstance(dtype, dt.Interval)
    assert dtype.is_interval() is True


@h.given(its.temporal_dtypes())
def test_temporal_dtype(dtype):
    assert isinstance(dtype, dt.Temporal)
    assert dtype.is_temporal() is True


@h.given(its.primitive_dtypes())
def test_primitive_dtype(dtype):
    assert isinstance(dtype, dt.Primitive)
    assert dtype.is_primitive() is True


@h.given(its.geospatial_dtypes())
def test_geospatial_dtype(dtype):
    assert isinstance(dtype, dt.GeoSpatial)
    assert dtype.is_geospatial() is True


@h.given(its.array_dtypes(its.primitive_dtypes()))
def test_array_dtype(dtype):
    assert isinstance(dtype, dt.Array)
    assert isinstance(dtype.value_type, dt.Primitive)
    assert dtype.is_array() is True


@h.given(its.array_dtypes(its.array_dtypes(its.primitive_dtypes())))
def test_array_array_dtype(dtype):
    assert isinstance(dtype, dt.Array)
    assert isinstance(dtype.value_type, dt.Array)
    assert isinstance(dtype.value_type.value_type, dt.Primitive)


@h.given(its.map_dtypes(its.primitive_dtypes(), its.boolean_dtype()))
def test_map_dtype(dtype):
    assert isinstance(dtype, dt.Map)
    assert isinstance(dtype.key_type, dt.Primitive)
    assert isinstance(dtype.value_type, dt.Boolean)
    assert dtype.is_map() is True


@h.given(its.struct_dtypes())
def test_struct_dtype(dtype):
    assert isinstance(dtype, dt.Struct)
    assert all(t.is_primitive() for t in dtype.types)
    assert dtype.is_struct() is True


@h.given(its.struct_dtypes(its.variadic_dtypes()))
def test_struct_variadic_dtype(dtype):
    assert isinstance(dtype, dt.Struct)
    assert all(t.is_variadic() for t in dtype.types)
    assert dtype.is_struct() is True


@h.given(its.variadic_dtypes())
def test_variadic_dtype(dtype):
    assert isinstance(dtype, dt.Variadic)
    assert dtype.is_variadic() is True


@h.given(its.all_dtypes())
def test_all_dtypes(dtype):
    assert isinstance(dtype, dt.DataType)


@h.given(its.schema())
def test_schema(schema):
    assert isinstance(schema, sch.Schema)
    assert all(t.is_primitive() for t in schema.types)
    assert all(isinstance(n, str) for n in schema.names)
    assert len(set(schema.names)) == len(schema.names)


@h.given(its.schema(its.array_dtypes(its.numeric_dtypes())))
def test_schema_array_dtype(schema):
    assert isinstance(schema, sch.Schema)
    assert all(t.is_array() for t in schema.types)
    assert all(isinstance(n, str) for n in schema.names)


@h.given(its.primitive_dtypes())
def test_primitive_dtypes_to_pandas(dtype):
    assert isinstance(dtype.to_pandas(), np.dtype)


@h.given(its.schema())
def test_schema_to_pandas(schema):
    pandas_schema = schema.to_pandas()
    assert len(pandas_schema) == len(schema)


@h.given(its.memtable(its.schema(its.integer_dtypes(), max_size=5)))
def test_memtable(memtable):
    assert isinstance(memtable, ir.Table)
    assert isinstance(memtable.schema(), sch.Schema)


@h.given(its.all_dtypes())
def test_deferred_literal(dtype):
    with pytest.raises(ValidationError):
        ibis.literal(ibis._.a, type=dtype)


# TODO(kszucs): we enforce field name uniqueness in the schema, but we don't for Struct datatype



from __future__ import annotations

from abc import abstractmethod
from typing import Generic, Optional

from public import public
from typing_extensions import Any, Self, TypeVar

import ibis.expr.datashape as ds
import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
from ibis import util
from ibis.common.annotations import attribute
from ibis.common.graph import Node as Traversable
from ibis.common.grounds import Concrete
from ibis.common.patterns import Coercible, CoercionError
from ibis.common.typing import DefaultTypeVars


@public
class Node(Concrete, Traversable):
    def equals(self, other) -> bool:
        if not isinstance(other, Node):
            raise TypeError(
                f"invalid equality comparison between Node and {type(other)}"
            )
        return self.__cached_equals__(other)

    @util.deprecated(as_of="4.0", instead="remove intermediate .op() calls")
    def op(self) -> Self:
        """Make `Node` backwards compatible with code that uses `Expr.op()`."""
        return self

    # Avoid custom repr for performance reasons
    __repr__ = object.__repr__

    # TODO(kszucs): hidrate the __children__ traversable attribute
    # @attribute
    # def __children__(self):
    #     return super().__children__


T = TypeVar("T", bound=dt.DataType, covariant=True)
S = TypeVar("S", bound=ds.DataShape, default=ds.Any, covariant=True)


@public
class Value(Node, Coercible, DefaultTypeVars, Generic[T, S]):
    @classmethod
    def __coerce__(
        cls, value: Any, T: Optional[type] = None, S: Optional[type] = None
    ) -> Self:
        # note that S=Shape is unused here since the pattern will check the
        # shape of the value expression after executing Value.__coerce__()
        from ibis.expr.operations.generic import NULL, Literal
        from ibis.expr.types import Expr

        if isinstance(value, Expr):
            value = value.op()

        if isinstance(value, Value):
            if value == NULL:
                # treat the NULL literal the same as None to implicitly cast to
                # the requested datatype if any
                value = None
            else:
                return value

        if T is dt.Integer:
            dtype = dt.infer(int(value))
        elif T is dt.Floating:
            dtype = dt.infer(float(value))
        else:
            try:
                dtype = dt.DataType.from_typehint(T)
            except TypeError:
                dtype = dt.infer(value)

        try:
            return Literal(value, dtype=dtype)
        except TypeError:
            raise CoercionError(f"Unable to coerce {value!r} to Value[{T!r}]")

    # TODO(kszucs): cover it with tests
    # TODO(kszucs): figure out how to represent not named arguments
    @property
    def name(self) -> str:
        names = (arg.name for arg in self.__args__ if hasattr(arg, "name"))
        return f"{self.__class__.__name__}({', '.join(names)})"

    @property
    @abstractmethod
    def dtype(self) -> T:
        """Ibis datatype of the produced value expression.

        Returns
        -------
        dt.DataType

        """

    @property
    @abstractmethod
    def shape(self) -> S:
        """Shape of the produced value expression.

        Possible values are: "scalar" and "columnar"

        Returns
        -------
        ds.Shape

        """

    @attribute
    def relations(self):
        """Set of relations the value node depends on."""
        children = (n.relations for n in self.__children__ if isinstance(n, Value))
        return frozenset().union(*children)

    @property
    @util.deprecated(as_of="7.0", instead="use .dtype property instead")
    def output_dtype(self):
        return self.dtype

    @property
    @util.deprecated(as_of="7.0", instead="use .shape property instead")
    def output_shape(self):
        return self.shape

    def to_expr(self):
        import ibis.expr.types as ir

        if self.shape.is_columnar():
            typename = self.dtype.column
        else:
            typename = self.dtype.scalar

        return getattr(ir, typename)(self)


# convenience aliases
Scalar = Value[T, ds.Scalar]
Column = Value[T, ds.Columnar]


@public
class Alias(Value):
    arg: Value
    name: str

    shape = rlz.shape_like("arg")
    dtype = rlz.dtype_like("arg")


@public
class Unary(Value):
    """A unary operation."""

    arg: Value

    @attribute
    def shape(self) -> ds.DataShape:
        return self.arg.shape

    @attribute
    def relations(self):
        return self.arg.relations


@public
class Binary(Value):
    """A binary operation."""

    left: Value
    right: Value

    @attribute
    def shape(self) -> ds.DataShape:
        return max(self.left.shape, self.right.shape)

    @attribute
    def relations(self):
        return self.left.relations | self.right.relations


@public
class Argument(Value):
    name: str
    shape: ds.DataShape
    dtype: dt.DataType

    @attribute
    def param(self) -> str:
        return f"__ibis_param_{self.name}__"


public(ValueOp=Value, UnaryOp=Unary, BinaryOp=Binary, Scalar=Scalar, Column=Column)



"""Lower the ibis expression graph to a SQL-like relational algebra."""


from __future__ import annotations

import operator
from functools import reduce
from typing import TYPE_CHECKING, Any, Literal, Optional

import toolz
from public import public

import ibis.common.exceptions as com
import ibis.expr.datashape as ds
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.common.annotations import attribute
from ibis.common.collections import FrozenDict  # noqa: TCH001
from ibis.common.deferred import var
from ibis.common.graph import Graph
from ibis.common.patterns import Object, Pattern, _, replace
from ibis.common.typing import VarTuple  # noqa: TCH001
from ibis.expr.rewrites import d, p, replace_parameter
from ibis.expr.schema import Schema

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

x = var("x")
y = var("y")


@public
class CTE(ops.Relation):
    """Common table expression."""

    parent: ops.Relation

    @attribute
    def schema(self):
        return self.parent.schema

    @attribute
    def values(self):
        return self.parent.values


@public
class Select(ops.Relation):
    """Relation modelled after SQL's SELECT statement."""

    parent: ops.Relation
    selections: FrozenDict[str, ops.Value] = {}
    predicates: VarTuple[ops.Value[dt.Boolean]] = ()
    sort_keys: VarTuple[ops.SortKey] = ()

    @attribute
    def values(self):
        return self.selections

    @attribute
    def schema(self):
        return Schema({k: v.dtype for k, v in self.selections.items()})


@public
class Window(ops.Value):
    """Window modelled after SQL's window statements."""

    how: Literal["rows", "range"]
    func: ops.Reduction | ops.Analytic
    start: Optional[ops.WindowBoundary] = None
    end: Optional[ops.WindowBoundary] = None
    group_by: VarTuple[ops.Column] = ()
    order_by: VarTuple[ops.SortKey] = ()

    shape = ds.columnar

    @attribute
    def dtype(self):
        return self.func.dtype


@replace(p.Project)
def project_to_select(_, **kwargs):
    """Convert a Project node to a Select node."""
    return Select(_.parent, selections=_.values)


@replace(p.Filter)
def filter_to_select(_, **kwargs):
    """Convert a Filter node to a Select node."""
    return Select(_.parent, selections=_.values, predicates=_.predicates)


@replace(p.Sort)
def sort_to_select(_, **kwargs):
    """Convert a Sort node to a Select node."""
    return Select(_.parent, selections=_.values, sort_keys=_.keys)


@replace(p.WindowFunction)
def window_function_to_window(_, **kwargs):
    """Convert a WindowFunction node to a Window node."""
    if isinstance(_.frame, ops.RowsWindowFrame) and _.frame.max_lookback is not None:
        raise NotImplementedError("max_lookback is not supported for SQL backends")
    return Window(
        how=_.frame.how,
        func=_.func,
        start=_.frame.start,
        end=_.frame.end,
        group_by=_.frame.group_by,
        order_by=_.frame.order_by,
    )


@replace(Object(Select, Object(Select)))
def merge_select_select(_, **kwargs):
    """Merge subsequent Select relations into one.

    This rewrites eliminates `_.parent` by merging the outer and the inner
    `predicates`, `sort_keys` and keeping the outer `selections`. All selections
    from the inner Select are inlined into the outer Select.
    """
    # don't merge if either the outer or the inner select has window functions
    for v in _.selections.values():
        if v.find(Window, filter=ops.Value):
            return _
    for v in _.parent.selections.values():
        if v.find((Window, ops.Unnest), filter=ops.Value):
            return _
    for v in _.predicates:
        if v.find((ops.ExistsSubquery, ops.InSubquery), filter=ops.Value):
            return _

    subs = {ops.Field(_.parent, k): v for k, v in _.parent.values.items()}
    selections = {k: v.replace(subs, filter=ops.Value) for k, v in _.selections.items()}
    predicates = tuple(p.replace(subs, filter=ops.Value) for p in _.predicates)
    sort_keys = tuple(s.replace(subs, filter=ops.Value) for s in _.sort_keys)

    return Select(
        _.parent.parent,
        selections=selections,
        predicates=tuple(toolz.unique(_.parent.predicates + predicates)),
        sort_keys=tuple(toolz.unique(_.parent.sort_keys + sort_keys)),
    )


def extract_ctes(node):
    result = []
    cte_types = (Select, ops.Aggregate, ops.JoinChain, ops.Set, ops.Limit, ops.Sample)

    g = Graph.from_bfs(node, filter=(ops.Relation, ops.Subquery, ops.JoinLink))
    for node, dependents in g.invert().items():
        if isinstance(node, ops.View) or (
            len(dependents) > 1 and isinstance(node, cte_types)
        ):
            result.append(node)

    return result


def sqlize(
    node: ops.Node,
    params: Mapping[ops.ScalarParameter, Any],
    rewrites: Sequence[Pattern] = (),
) -> tuple[ops.Node, list[ops.Node]]:
    """Lower the ibis expression graph to a SQL-like relational algebra.

    Parameters
    ----------
    node
        The root node of the expression graph.
    params
        A mapping of scalar parameters to their values.
    rewrites
        Supplementary rewrites to apply to the expression graph.

    Returns
    -------
    Tuple of the rewritten expression graph and a list of CTEs.

    """
    assert isinstance(node, ops.Relation)

    # apply the backend specific rewrites
    node = node.replace(reduce(operator.or_, rewrites))

    # lower the expression graph to a SQL-like relational algebra
    context = {"params": params}
    sqlized = node.replace(
        replace_parameter
        | project_to_select
        | filter_to_select
        | sort_to_select
        | window_function_to_window,
        context=context,
    )

    # squash subsequent Select nodes into one
    simplified = sqlized.replace(merge_select_select)

    # extract common table expressions while wrapping them in a CTE node
    ctes = extract_ctes(simplified)
    subs = {cte: CTE(cte) for cte in ctes}
    result = simplified.replace(subs)

    return result, ctes


# supplemental rewrites selectively used on a per-backend basis

"""Replace `log2` and `log10` with `log`."""
replace_log2 = p.Log2 >> d.Log(_.arg, base=2)
replace_log10 = p.Log10 >> d.Log(_.arg, base=10)


"""Add an ORDER BY clause to rank window functions that don't have one."""
add_order_by_to_empty_ranking_window_functions = p.WindowFunction(
    func=p.NTile(y),
    frame=p.WindowFrame(order_by=()) >> _.copy(order_by=(y,)),
)

"""Replace checks against an empty right side with `False`."""
empty_in_values_right_side = p.InValues(options=()) >> d.Literal(False, dtype=dt.bool)


@replace(
    p.WindowFunction(p.RankBase | p.NTile)
    | p.StringFind
    | p.FindInSet
    | p.ArrayPosition
)
def one_to_zero_index(_, **kwargs):
    """Subtract one from one-index functions."""
    return ops.Subtract(_, 1)


@replace(ops.NthValue)
def add_one_to_nth_value_input(_, **kwargs):
    if isinstance(_.nth, ops.Literal):
        nth = ops.Literal(_.nth.value + 1, dtype=_.nth.dtype)
    else:
        nth = ops.Add(_.nth, 1)
    return _.copy(nth=nth)


@replace(p.Capitalize)
def rewrite_capitalize(_, **kwargs):
    """Rewrite Capitalize in terms of substring, concat, upper, and lower."""
    first = ops.Uppercase(ops.Substring(_.arg, start=0, length=1))
    # use length instead of length - 1 to avoid backends complaining about
    # asking for negative length
    #
    # there are at most length - 1 characters, so asking for length is fine
    rest = ops.Lowercase(ops.Substring(_.arg, start=1, length=ops.StringLength(_.arg)))
    return ops.StringConcat((first, rest))


@replace(p.Sample)
def rewrite_sample_as_filter(_, **kwargs):
    """Rewrite Sample as `t.filter(random() <= fraction)`.

    Errors as unsupported if a `seed` is specified.
    """
    if _.seed is not None:
        raise com.UnsupportedOperationError(
            "`Table.sample` with a random seed is unsupported"
        )
    return ops.Filter(_.parent, (ops.LessEqual(ops.RandomScalar(), _.fraction),))


@replace(p.WindowFunction(p.First(x, where=y)))
def rewrite_first_to_first_value(_, x, y, **kwargs):
    """Rewrite Ibis's first to first_value when used in a window function."""
    if y is not None:
        raise com.UnsupportedOperationError(
            "`first` with `where` is unsupported in a window function"
        )
    return _.copy(func=ops.FirstValue(x))


@replace(p.WindowFunction(p.Last(x, where=y)))
def rewrite_last_to_last_value(_, x, y, **kwargs):
    """Rewrite Ibis's last to last_value when used in a window function."""
    if y is not None:
        raise com.UnsupportedOperationError(
            "`last` with `where` is unsupported in a window function"
        )
    return _.copy(func=ops.LastValue(x))


@replace(p.WindowFunction(frame=y @ p.WindowFrame(order_by=())))
def rewrite_empty_order_by_window(_, y, **kwargs):
    return _.copy(frame=y.copy(order_by=(ops.NULL,)))


@replace(p.WindowFunction(p.RowNumber | p.NTile, y))
def exclude_unsupported_window_frame_from_row_number(_, y):
    return ops.Subtract(_.copy(frame=y.copy(start=None, end=0)), 1)


@replace(p.WindowFunction(p.MinRank | p.DenseRank, y @ p.WindowFrame(start=None)))
def exclude_unsupported_window_frame_from_rank(_, y):
    return ops.Subtract(
        _.copy(frame=y.copy(start=None, end=0, order_by=y.order_by or (ops.NULL,))), 1
    )


@replace(
    p.WindowFunction(
        p.Lag | p.Lead | p.PercentRank | p.CumeDist | p.Any | p.All,
        y @ p.WindowFrame(start=None),
    )
)
def exclude_unsupported_window_frame_from_ops(_, y, **kwargs):
    return _.copy(frame=y.copy(start=None, end=0, order_by=y.order_by or (ops.NULL,)))



# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import pytest

import ibis
import ibis.expr.types as ir
from ibis.common.annotations import ValidationError
from ibis.tests.expr.mocks import MockBackend
from ibis.tests.util import assert_equal


@pytest.fixture
def con():
    return MockBackend()


@pytest.fixture
def alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture
def airlines():
    return ibis.table(
        [("dest", "string"), ("origin", "string"), ("arrdelay", "int32")],
        "airlines",
    )


def test_category_project(alltypes):
    t = alltypes

    tier = t.double_col.bucket([0, 50, 100]).name("tier")
    expr = t[tier, t]

    assert isinstance(expr.tier, ir.IntegerColumn)


def test_bucket(alltypes):
    d = alltypes.double_col
    bins = [0, 10, 50, 100]

    expr = d.bucket(bins)
    assert isinstance(expr, ir.IntegerColumn)
    assert expr.op().nbuckets == 3

    expr = d.bucket(bins, include_over=True)
    assert expr.op().nbuckets == 4

    expr = d.bucket(bins, include_over=True, include_under=True)
    assert expr.op().nbuckets == 5


def test_bucket_error_cases(alltypes):
    d = alltypes.double_col

    with pytest.raises(ValidationError):
        d.bucket([])

    with pytest.raises(ValidationError):
        d.bucket([1, 2], closed="foo")

    # it works!
    d.bucket([10], include_under=True, include_over=True)

    with pytest.raises(ValidationError):
        d.bucket([10])

    with pytest.raises(ValidationError):
        d.bucket([10], include_under=True)

    with pytest.raises(ValidationError):
        d.bucket([10], include_over=True)


def test_histogram(alltypes):
    d = alltypes.double_col

    with pytest.raises(ValueError):
        d.histogram(nbins=10, binwidth=5)

    with pytest.raises(ValueError):
        d.histogram()


def test_topk_analysis_bug(airlines):
    # GH #398
    dests = ["ORD", "JFK", "SFO"]
    t = airlines[airlines.dest.isin(dests)]
    filtered = t.semi_join(t.origin.topk(10, by=t.arrdelay.mean()), "origin")
    assert filtered is not None


def test_topk_function_late_bind(airlines):
    # GH #520
    expr1 = airlines.dest.topk(5, by=lambda x: x.arrdelay.mean())
    expr2 = airlines.dest.topk(5, by=airlines.arrdelay.mean())

    assert_equal(expr1, expr2)



from __future__ import annotations

import pandas as pd
import pytest
from dask.dataframe.utils import tm

import ibis

dd = pytest.importorskip("dask.dataframe")


def test_table_from_dataframe(dataframe, ibis_table, con):
    t = con.from_dataframe(dataframe)
    result = t.execute()
    expected = ibis_table.execute()
    tm.assert_frame_equal(result, expected)

    t = con.from_dataframe(dataframe, name="foo")
    expected = ibis_table.execute()
    tm.assert_frame_equal(result, expected)

    t = con.from_dataframe(dataframe, name="foo", client=con)
    expected = ibis_table.execute()
    tm.assert_frame_equal(result, expected)


def test_array_literal_from_series(con):
    values = [1, 2, 3, 4]
    s = dd.from_pandas(pd.Series(values), npartitions=1)
    expr = ibis.array(s)

    assert expr.equals(ibis.array(values))
    assert con.execute(expr) == pytest.approx([1, 2, 3, 4])


def test_execute_parameter_only(con):
    param = ibis.param("int64")
    result = con.execute(param, params={param.op(): 42})
    assert result == 42



from __future__ import annotations

import datetime
import functools
import re
from pathlib import Path
from typing import TYPE_CHECKING, Callable

import pytest
import sqlglot as sg
from dateutil.relativedelta import relativedelta

import ibis
from ibis.formats.pandas import PandasData

if TYPE_CHECKING:
    import ibis.expr.types as ir


def pytest_pyfunc_call(pyfuncitem):
    """Inject `backend` and `snapshot` fixtures to all TPC-H test functions.

    Defining this hook here limits its scope to the TPC-H tests.
    """
    testfunction = pyfuncitem.obj
    funcargs = pyfuncitem.funcargs
    testargs = {arg: funcargs[arg] for arg in pyfuncitem._fixtureinfo.argnames}
    result = testfunction(
        **testargs, backend=funcargs["backend"], snapshot=funcargs["snapshot"]
    )
    assert (
        result is None
    ), "test function should not return anything, did you mean to use assert?"
    return True


def tpch_test(test: Callable[..., ir.Table]):
    """Decorator for TPCH tests.

    Automates the process of loading the SQL query from the file system and
    asserting that the result of the ibis expression is equal to the expected
    result of executing the raw SQL.
    """

    @pytest.mark.tpch
    @pytest.mark.usefixtures("backend", "snapshot")
    @pytest.mark.xdist_group("tpch")
    @functools.wraps(test)
    def wrapper(*args, backend, snapshot, **kwargs):
        backend_name = backend.name()
        if not backend.supports_tpch:
            pytest.skip(
                f"{backend_name} backend doesn't support testing TPC-H queries yet"
            )
        query_name_match = re.match(r"^test_tpc_(h\d\d)$", test.__name__)
        assert query_name_match is not None

        query_number = query_name_match.group(1)
        sql_path_name = f"{query_number}.sql"

        path = Path(__file__).parent.joinpath("queries", "duckdb", sql_path_name)
        raw_sql = path.read_text()

        sql = sg.parse_one(raw_sql, read="duckdb")

        sql = backend._transform_tpch_sql(sql)

        raw_sql = sql.sql(dialect="duckdb", pretty=True)

        expected_expr = backend.connection.sql(raw_sql, dialect="duckdb")

        result_expr = test(*args, **kwargs)

        ibis_sql = ibis.to_sql(result_expr, dialect=backend_name)

        assert result_expr._find_backend(use_default=False) is backend.connection
        result = backend.connection.execute(result_expr)
        assert not result.empty

        expected = expected_expr.execute()
        assert list(map(str.lower, expected.columns)) == result.columns.tolist()
        expected.columns = result.columns

        expected = PandasData.convert_table(expected, result_expr.schema())
        assert not expected.empty

        assert len(expected) == len(result)
        assert result.columns.tolist() == expected.columns.tolist()
        for column in result.columns:
            left = result.loc[:, column]
            right = expected.loc[:, column]
            assert (
                pytest.approx(left.values.tolist(), nan_ok=True)
                == right.values.tolist()
            )

        # only write sql if the execution passes
        snapshot.assert_match(ibis_sql, sql_path_name)

    return wrapper


def add_date(datestr: str, dy: int = 0, dm: int = 0, dd: int = 0) -> ir.DateScalar:
    dt = datetime.date.fromisoformat(datestr)
    dt += relativedelta(years=dy, months=dm, days=dd)
    return ibis.date(dt.isoformat())


@pytest.fixture(scope="session")
def customer(backend):
    return backend.customer


@pytest.fixture(scope="session")
def lineitem(backend):
    return backend.lineitem


@pytest.fixture(scope="session")
def nation(backend):
    return backend.nation


@pytest.fixture(scope="session")
def orders(backend):
    return backend.orders


@pytest.fixture(scope="session")
def part(backend):
    return backend.part


@pytest.fixture(scope="session")
def partsupp(backend):
    return backend.partsupp


@pytest.fixture(scope="session")
def region(backend):
    return backend.region


@pytest.fixture(scope="session")
def supplier(backend):
    return backend.supplier



from __future__ import annotations

from functools import partial
from typing import Union

import pytest

import ibis.expr.datashape as ds
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.common.annotations import ValidationError
from ibis.common.patterns import NoMatch, Pattern

one = ops.Literal(1, dt.int8)


@pytest.mark.parametrize(
    ("value", "dtype"),
    [
        (1, dt.Int8),
        (1.0, dt.Float64),
        (True, dt.Boolean),
        ("foo", dt.String),
        (b"foo", dt.Binary),
        ((1, 2), dt.Array[dt.Int8]),
    ],
)
def test_literal_coercion_type_inference(value, dtype):
    assert ops.Literal.__coerce__(value) == ops.Literal(value, dtype)
    assert ops.Literal.__coerce__(value, dtype) == ops.Literal(value, dtype)


@pytest.mark.parametrize(
    ("typehint", "value", "expected"),
    [
        (ops.Literal, 1, one),
        (ops.Literal, one, one),
        (ops.Literal, False, ops.Literal(False, dt.boolean)),
        (ops.Literal[dt.Int8], 1, one),
        (ops.Literal[dt.Int16], 1, ops.Literal(1, dt.int16)),
        (ops.Literal[dt.Int8], ops.Literal(1, dt.int16), NoMatch),
        (ops.Literal[dt.Integer], 1, ops.Literal(1, dt.int8)),
        (ops.Literal[dt.Floating], 1, ops.Literal(1, dt.float64)),
        (ops.Literal[dt.Float32], 1.0, ops.Literal(1.0, dt.float32)),
    ],
)
def test_coerced_to_literal(typehint, value, expected):
    pat = Pattern.from_typehint(typehint)
    assert pat.match(value, {}) == expected


@pytest.mark.parametrize(
    ("typehint", "value", "expected"),
    [
        (ops.Value, 1, one),
        (ops.Value[dt.Int8], 1, one),
        (ops.Value[dt.Int8, ds.Any], 1, one),
        (ops.Value[dt.Int8, ds.Scalar], 1, one),
        (ops.Value[dt.Int8, ds.Columnar], 1, NoMatch),
        # dt.Integer is not instantiable so it will be only used for checking
        # that the produced literal has any integer datatype
        (ops.Value[dt.Integer], 1, one),
        # same applies here, the coercion itself will use only the inferred datatype
        # but then the result is checked against the given typehint
        (ops.Value[dt.Int8 | dt.Int16], 1, one),
        (Union[ops.Value[dt.Int8], ops.Value[dt.Int16]], 1, one),
        (ops.Value[dt.Int8 | dt.Int16], 128, ops.Literal(128, dt.int16)),
        (
            Union[ops.Value[dt.Int8], ops.Value[dt.Int16]],
            128,
            ops.Literal(128, dt.int16),
        ),
        (ops.Value[dt.Int8 | dt.Int16], 128, ops.Literal(128, dt.int16)),
        (
            Union[ops.Value[dt.Int8], ops.Value[dt.Int16]],
            128,
            ops.Literal(128, dt.int16),
        ),
        (ops.Value[dt.Int8], 128, NoMatch),
        # this is actually supported by creating an explicit dtype
        # in Value.__coerce__ based on the `T` keyword argument
        (ops.Value[dt.Int16, ds.Scalar], 1, ops.Literal(1, dt.int16)),
        (ops.Value[dt.Int16, ds.Scalar], 128, ops.Literal(128, dt.int16)),
        # equivalent with ops.Value[dt.Int8 | dt.Int16]
        (Union[ops.Value[dt.Int8], ops.Value[dt.Int16]], 1, one),
        # when expecting floating point values given an integer value it will
        # be coerced to float64
        (ops.Value[dt.Floating], 1, ops.Literal(1, dt.float64)),
    ],
)
def test_coerced_to_value(typehint, value, expected):
    pat = Pattern.from_typehint(typehint)
    assert pat.match(value, {}) == expected


@pytest.mark.pandas
def test_coerced_to_interval_value():
    import pandas as pd

    expected = ops.Literal(1, dt.Interval("s"))
    pat = Pattern.from_typehint(ops.Value[dt.Interval])
    assert pat.match(pd.Timedelta("1s"), {}) == expected

    expected = ops.Literal(3661, dt.Interval("s"))
    assert pat.match(pd.Timedelta("1h 1m 1s"), {}) == expected


@pytest.mark.parametrize(
    ("call", "error"),
    [
        (partial(ops.Literal, 1), "missing_a_required_argument"),
        (partial(ops.Literal, 1, dt.int8, "foo"), "too_many_positional_arguments"),
        (partial(ops.Literal, 1, dt.int8, name="foo"), "got_an_unexpected_keyword"),
        (
            partial(ops.Literal, 1, dt.int8, dtype=dt.int16),
            "multiple_values_for_argument",
        ),
        (partial(ops.Literal, 1, 4), "invalid_dtype"),
    ],
)
def test_error_message_when_constructing_literal(call, error, snapshot):
    with pytest.raises(ValidationError) as exc:
        call()
    snapshot.assert_match(str(exc.value), f"{error}.txt")


def test_implicit_coercion_of_null_literal():
    # GH #7775
    NULL = ops.Literal(None, dt.null)

    value = ops.Value.__coerce__(None, dt.Int8)
    expected = ops.Literal(None, dt.int8)
    assert value == expected

    value = ops.Value.__coerce__(NULL, dt.Float64)
    expected = ops.Literal(None, dt.float64)
    assert value == expected


def test_NULL():
    assert isinstance(ops.NULL, ops.Literal)
    assert ops.NULL.value is None
    assert ops.NULL.dtype is dt.null
    assert ops.NULL == ops.Literal(None, dt.null)
    assert ops.NULL is not ops.Literal(None, dt.int8)



from __future__ import annotations

import pandas as pd
import pytest
from pytest import param

import ibis
import ibis.common.exceptions as com
from ibis import literal as L
from ibis.backends.impala.tests.conftest import translate


@pytest.fixture(scope="module")
def table(mockcon):
    return mockcon.table("alltypes")


@pytest.mark.parametrize(
    "value",
    [
        param("simple", id="simple"),
        param("I can't", id="embedded_single_quote"),
        param('An "escape"', id="embedded_double_quote"),
        param(5, id="int"),
        param(1.5, id="float"),
        param(True, id="true"),
        param(False, id="false"),
    ],
)
def test_literals(value, snapshot):
    expr = L(value)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


def test_column_ref_table_aliases():
    table1 = ibis.table([("key1", "string"), ("value1", "double")])
    table2 = ibis.table([("key2", "string"), ("value and2", "double")])

    expr = table1["value1"] - table2["value and2"]

    with pytest.raises(com.RelationError, match="multiple base table references"):
        translate(expr)


@pytest.mark.parametrize(
    "expr_fn",
    [
        lambda t: t.g.cast("double").name("g_dub"),
        lambda t: t.g.name("has a space"),
        lambda t: ((t.a - t.b) * t.a).name("expr"),
    ],
    ids=["cast", "spaces", "compound_expr"],
)
def test_named_expressions(table, expr_fn, snapshot):
    expr = expr_fn(table)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "expr_fn",
    [
        param(lambda t: t.a + t.b, id="add"),
        param(lambda t: t.a - t.b, id="sub"),
        param(lambda t: t.a * t.b, id="mul"),
        param(lambda t: t.a / t.b, id="div"),
        param(lambda t: t.a**t.b, id="pow"),
        param(lambda t: t.a < t.b, id="lt"),
        param(lambda t: t.a <= t.b, id="le"),
        param(lambda t: t.a > t.b, id="gt"),
        param(lambda t: t.a >= t.b, id="ge"),
        param(lambda t: t.a == t.b, id="eq"),
        param(lambda t: t.a != t.b, id="ne"),
        param(lambda t: t.h & (t.a > 0), id="and"),
        param(lambda t: t.h | (t.a > 0), id="or"),
        param(lambda t: t.h ^ (t.a > 0), id="xor"),
    ],
)
def test_binary_infix_operators(table, expr_fn, snapshot):
    expr = expr_fn(table)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "expr_fn",
    [
        lambda t: (t.a + t.b) + t.c,
        lambda t: t.a.log() + t.c,
        lambda t: t.b + (-(t.a + t.c)),
    ],
    ids=["parens_left", "function_call", "negation"],
)
def test_binary_infix_parenthesization(table, expr_fn, snapshot):
    expr = expr_fn(table)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


def test_between(table, snapshot):
    expr = table.f.between(0, 1)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "expr_fn",
    [
        lambda t: t["g"].isnull(),
        lambda t: t["a"].notnull(),
        lambda t: (t["a"] + t["b"]).isnull(),
    ],
    ids=["isnull", "notnull", "compound_isnull"],
)
def test_isnull_notnull(table, expr_fn, snapshot):
    expr = expr_fn(table)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    ("column", "to_type"),
    [
        ("a", "int16"),
        ("a", "int32"),
        ("a", "int64"),
        ("a", "string"),
        ("d", "int8"),
        ("g", "double"),
        ("g", "timestamp"),
    ],
)
def test_casts(table, column, to_type, snapshot):
    expr = table[column].cast(to_type)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


def test_misc_conditionals(table, snapshot):
    expr = table.a.nullif(0)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "expr_fn",
    [
        lambda _: L("9.9999999").cast("decimal(38, 5)"),
        lambda t: t.f.cast("decimal(12, 2)"),
    ],
    ids=["literal", "column"],
)
def test_decimal_casts(table, expr_fn, snapshot):
    expr = expr_fn(table)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize("colname", ["a", "f", "h"])
def test_negate(table, colname, snapshot):
    result = translate(-table[colname])
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "field",
    ["year", "month", "day", "hour", "minute", "second", "microsecond", "millisecond"],
)
def test_timestamp_extract_field(table, field, snapshot):
    expr = getattr(table.i, field)()
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


def test_sql_extract(table, snapshot):
    # integration with SQL translation
    expr = table[
        table.i.year().name("year"),
        table.i.month().name("month"),
        table.i.day().name("day"),
    ]

    result = ibis.to_sql(expr, dialect="impala")
    snapshot.assert_match(result, "out.sql")


def test_timestamp_now(snapshot):
    expr = ibis.now()
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "unit",
    ["years", "months", "weeks", "days", "hours", "minutes", "seconds"],
)
def test_timestamp_deltas(table, unit, snapshot):
    K = 5

    offset = ibis.interval(**{unit: K})

    add_expr = table.i + offset
    result = translate(add_expr)
    snapshot.assert_match(result, "out1.sql")

    sub_expr = table.i - offset
    result = translate(sub_expr)
    snapshot.assert_match(result, "out2.sql")


@pytest.mark.parametrize(
    "expr_fn",
    [
        lambda v: L(pd.Timestamp(v)),
        lambda v: L(pd.Timestamp(v).to_pydatetime()),
        lambda v: ibis.timestamp(v),
    ],
    ids=["pd_timestamp", "pydatetime", "timestamp_function"],
)
def test_timestamp_literals(expr_fn, snapshot):
    expr = expr_fn("2015-01-01 12:34:56")
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize("method_name", ["index", "full_name"])
def test_timestamp_day_of_week(method_name, snapshot):
    ts = ibis.timestamp("2015-09-01T01:00:23")
    expr = getattr(ts.day_of_week, method_name)()
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "expr_fn",
    [
        lambda col: col.to_timestamp(),
        lambda col: col.to_timestamp("ms"),
        lambda col: col.to_timestamp("us"),
    ],
    ids=["default", "ms", "us"],
)
def test_timestamp_from_integer(table, expr_fn, snapshot):
    expr = expr_fn(table.c)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


def test_correlated_predicate_subquery(table, snapshot):
    t0 = table
    t1 = t0.view()

    # both are valid constructions
    expr1 = t0[t0.g == t1.g]
    expr2 = t1[t0.g == t1.g]

    snapshot.assert_match(translate(expr1), "out1.sql")
    snapshot.assert_match(translate(expr2), "out2.sql")


@pytest.mark.parametrize(
    "expr_fn",
    [
        param(lambda b: b.any(), id="any"),
        param(lambda b: -b.any(), id="not_any"),
        param(lambda b: b.all(), id="all"),
        param(lambda b: -b.all(), id="not_all"),
    ],
)
def test_any_all(table, expr_fn, snapshot):
    expr = expr_fn(table.f == 0)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")



from __future__ import annotations

from typing import Optional

from public import public

import ibis.expr.datatypes as dt
from ibis.common.annotations import attribute
from ibis.expr.operations.core import Column, Scalar  # noqa: TCH001
from ibis.expr.operations.relations import Relation
from ibis.expr.schema import Schema


@public
class WindowingTVF(Relation):
    """Generic windowing table-valued function."""

    # TODO(kszucs): rename to `parent`
    table: Relation
    time_col: Column[dt.Timestamp]  # enforce timestamp column type here

    @attribute
    def values(self):
        return self.table.fields

    @property
    def schema(self):
        names = list(self.table.schema.names)
        types = list(self.table.schema.types)

        # The return value of windowing TVF is a new relation that includes all columns
        # of original relation as well as additional 3 columns named “window_start”,
        # “window_end”, “window_time” to indicate the assigned window

        # TODO(kszucs): this looks like an implementation detail leaked from the
        # flink backend
        names.extend(["window_start", "window_end", "window_time"])
        # window_start, window_end, window_time have type TIMESTAMP(3) in Flink
        types.extend([dt.timestamp(scale=3)] * 3)

        return Schema.from_tuples(list(zip(names, types)))


@public
class TumbleWindowingTVF(WindowingTVF):
    """TUMBLE window table-valued function."""

    window_size: Scalar[dt.Interval]
    offset: Optional[Scalar[dt.Interval]] = None


@public
class HopWindowingTVF(WindowingTVF):
    """HOP window table-valued function."""

    window_size: Scalar[dt.Interval]
    window_slide: Scalar[dt.Interval]
    offset: Optional[Scalar[dt.Interval]] = None


@public
class CumulateWindowingTVF(WindowingTVF):
    """CUMULATE window table-valued function."""

    window_size: Scalar[dt.Interval]
    window_step: Scalar[dt.Interval]
    offset: Optional[Scalar[dt.Interval]] = None
